from __future__ import annotations

import pytest

import ibis

MOCK_TABLES = {
    "alltypes": [
        ("a", "int8"),
        ("b", "int16"),
        ("c", "int32"),
        ("d", "int64"),
        ("e", "float32"),
        ("f", "float64"),
        ("g", "string"),
        ("h", "boolean"),
        ("i", "timestamp"),
        ("j", "date"),
        ("k", "time"),
    ],
    "foo_t": [("key1", "string"), ("key2", "string"), ("value1", "double")],
    "bar_t": [("key1", "string"), ("key2", "string"), ("value2", "double")],
    "t1": [("key1", "string"), ("key2", "string"), ("value1", "double")],
    "t2": [("key1", "string"), ("key2", "string"), ("value2", "double")],
    "t3": [("key2", "string"), ("key3", "string"), ("value3", "double")],
    "t4": [("key3", "string"), ("value4", "double")],
    "bar": [("x", "double"), ("job", "string")],
    "foo": [
        ("job", "string"),
        ("dept_id", "string"),
        ("year", "int32"),
        ("y", "double"),
    ],
    "star1": [
        ("c", "int32"),
        ("f", "double"),
        ("foo_id", "string"),
        ("bar_id", "string"),
    ],
    "star2": [
        ("foo_id", "string"),
        ("value1", "double"),
        ("value3", "double"),
    ],
    "star3": [("bar_id", "string"), ("value2", "double")],
    "test1": [("c", "int32"), ("f", "double"), ("g", "string")],
    "test2": [("key", "string"), ("value", "double")],
    "geo": [
        ("id", "int32"),
        ("geo_point", "point"),
        ("geo_linestring", "linestring"),
        ("geo_polygon", "polygon"),
        ("geo_multipolygon", "multipolygon"),
    ],
    "tpch_region": [
        ("r_regionkey", "int16"),
        ("r_name", "string"),
        ("r_comment", "string"),
    ],
    "tpch_nation": [
        ("n_nationkey", "int16"),
        ("n_name", "string"),
        ("n_regionkey", "int16"),
        ("n_comment", "string"),
    ],
    "tpch_lineitem": [
        ("l_orderkey", "int64"),
        ("l_partkey", "int64"),
        ("l_suppkey", "int64"),
        ("l_linenumber", "int32"),
        ("l_quantity", "decimal(12,2)"),
        ("l_extendedprice", "decimal(12,2)"),
        ("l_discount", "decimal(12,2)"),
        ("l_tax", "decimal(12,2)"),
        ("l_returnflag", "string"),
        ("l_linestatus", "string"),
        ("l_shipdate", "string"),
        ("l_commitdate", "string"),
        ("l_receiptdate", "string"),
        ("l_shipinstruct", "string"),
        ("l_shipmode", "string"),
        ("l_comment", "string"),
    ],
    "tpch_customer": [
        ("c_custkey", "int64"),
        ("c_name", "string"),
        ("c_address", "string"),
        ("c_nationkey", "int16"),
        ("c_phone", "string"),
        ("c_acctbal", "decimal"),
        ("c_mktsegment", "string"),
        ("c_comment", "string"),
    ],
    "tpch_orders": [
        ("o_orderkey", "int64"),
        ("o_custkey", "int64"),
        ("o_orderstatus", "string"),
        ("o_totalprice", "decimal(12,2)"),
        ("o_orderdate", "string"),
        ("o_orderpriority", "string"),
        ("o_clerk", "string"),
        ("o_shippriority", "int32"),
        ("o_comment", "string"),
    ],
    "functional_alltypes": [
        ("id", "int32"),
        ("bool_col", "boolean"),
        ("tinyint_col", "int8"),
        ("smallint_col", "int16"),
        ("int_col", "int32"),
        ("bigint_col", "int64"),
        ("float_col", "float32"),
        ("double_col", "float64"),
        ("date_string_col", "string"),
        ("string_col", "string"),
        ("timestamp_col", "timestamp"),
        ("year", "int32"),
        ("month", "int32"),
    ],
    "airlines": [
        ("year", "int32"),
        ("month", "int32"),
        ("day", "int32"),
        ("dayofweek", "int32"),
        ("dep_time", "int32"),
        ("crs_dep_time", "int32"),
        ("arr_time", "int32"),
        ("crs_arr_time", "int32"),
        ("carrier", "string"),
        ("flight_num", "int32"),
        ("tail_num", "int32"),
        ("actual_elapsed_time", "int32"),
        ("crs_elapsed_time", "int32"),
        ("airtime", "int32"),
        ("arrdelay", "int32"),
        ("depdelay", "int32"),
        ("origin", "string"),
        ("dest", "string"),
        ("distance", "int32"),
        ("taxi_in", "int32"),
        ("taxi_out", "int32"),
        ("cancelled", "int32"),
        ("cancellation_code", "string"),
        ("diverted", "int32"),
        ("carrier_delay", "int32"),
        ("weather_delay", "int32"),
        ("nas_delay", "int32"),
        ("security_delay", "int32"),
        ("late_aircraft_delay", "int32"),
    ],
    "tpcds_customer": [
        ("c_customer_sk", "int64"),
        ("c_customer_id", "string"),
        ("c_current_cdemo_sk", "int32"),
        ("c_current_hdemo_sk", "int32"),
        ("c_current_addr_sk", "int32"),
        ("c_first_shipto_date_sk", "int32"),
        ("c_first_sales_date_sk", "int32"),
        ("c_salutation", "string"),
        ("c_first_name", "string"),
        ("c_last_name", "string"),
        ("c_preferred_cust_flag", "string"),
        ("c_birth_day", "int32"),
        ("c_birth_month", "int32"),
        ("c_birth_year", "int32"),
        ("c_birth_country", "string"),
        ("c_login", "string"),
        ("c_email_address", "string"),
        ("c_last_review_date", "string"),
    ],
    "tpcds_customer_address": [
        ("ca_address_sk", "bigint"),
        ("ca_address_id", "string"),
        ("ca_street_number", "string"),
        ("ca_street_name", "string"),
        ("ca_street_type", "string"),
        ("ca_suite_number", "string"),
        ("ca_city", "string"),
        ("ca_county", "string"),
        ("ca_state", "string"),
        ("ca_zip", "string"),
        ("ca_country", "string"),
        ("ca_gmt_offset", "decimal(5,2)"),
        ("ca_location_type", "string"),
    ],
    "tpcds_customer_demographics": [
        ("cd_demo_sk", "bigint"),
        ("cd_gender", "string"),
        ("cd_marital_status", "string"),
        ("cd_education_status", "string"),
        ("cd_purchase_estimate", "int"),
        ("cd_credit_rating", "string"),
        ("cd_dep_count", "int"),
        ("cd_dep_employed_count", "int"),
        ("cd_dep_college_count", "int"),
    ],
    "tpcds_date_dim": [
        ("d_date_sk", "bigint"),
        ("d_date_id", "string"),
        ("d_date", "string"),
        ("d_month_seq", "int"),
        ("d_week_seq", "int"),
        ("d_quarter_seq", "int"),
        ("d_year", "int"),
        ("d_dow", "int"),
        ("d_moy", "int"),
        ("d_dom", "int"),
        ("d_qoy", "int"),
        ("d_fy_year", "int"),
        ("d_fy_quarter_seq", "int"),
        ("d_fy_week_seq", "int"),
        ("d_day_name", "string"),
        ("d_quarter_name", "string"),
        ("d_holiday", "string"),
        ("d_weekend", "string"),
        ("d_following_holiday", "string"),
        ("d_first_dom", "int"),
        ("d_last_dom", "int"),
        ("d_same_day_ly", "int"),
        ("d_same_day_lq", "int"),
        ("d_current_day", "string"),
        ("d_current_week", "string"),
        ("d_current_month", "string"),
        ("d_current_quarter", "string"),
        ("d_current_year", "string"),
    ],
    "tpcds_household_demographics": [
        ("hd_demo_sk", "bigint"),
        ("hd_income_band_sk", "int"),
        ("hd_buy_potential", "string"),
        ("hd_dep_count", "int"),
        ("hd_vehicle_count", "int"),
    ],
    "tpcds_item": [
        ("i_item_sk", "bigint"),
        ("i_item_id", "string"),
        ("i_rec_start_date", "string"),
        ("i_rec_end_date", "string"),
        ("i_item_desc", "string"),
        ("i_current_price", "decimal(7,2)"),
        ("i_wholesale_cost", "decimal(7,2)"),
        ("i_brand_id", "int"),
        ("i_brand", "string"),
        ("i_class_id", "int"),
        ("i_class", "string"),
        ("i_category_id", "int"),
        ("i_category", "string"),
        ("i_manufact_id", "int"),
        ("i_manufact", "string"),
        ("i_size", "string"),
        ("i_formulation", "string"),
        ("i_color", "string"),
        ("i_units", "string"),
        ("i_container", "string"),
        ("i_manager_id", "int"),
        ("i_product_name", "string"),
    ],
    "tpcds_promotion": [
        ("p_promo_sk", "bigint"),
        ("p_promo_id", "string"),
        ("p_start_date_sk", "int"),
        ("p_end_date_sk", "int"),
        ("p_item_sk", "int"),
        ("p_cost", "decimal(15,2)"),
        ("p_response_target", "int"),
        ("p_promo_name", "string"),
        ("p_channel_dmail", "string"),
        ("p_channel_email", "string"),
        ("p_channel_catalog", "string"),
        ("p_channel_tv", "string"),
        ("p_channel_radio", "string"),
        ("p_channel_press", "string"),
        ("p_channel_event", "string"),
        ("p_channel_demo", "string"),
        ("p_channel_details", "string"),
        ("p_purpose", "string"),
        ("p_discount_active", "string"),
    ],
    "tpcds_store": [
        ("s_store_sk", "bigint"),
        ("s_store_id", "string"),
        ("s_rec_start_date", "string"),
        ("s_rec_end_date", "string"),
        ("s_closed_date_sk", "int"),
        ("s_store_name", "string"),
        ("s_number_employees", "int"),
        ("s_floor_space", "int"),
        ("s_hours", "string"),
        ("s_manager", "string"),
        ("s_market_id", "int"),
        ("s_geography_class", "string"),
        ("s_market_desc", "string"),
        ("s_market_manager", "string"),
        ("s_division_id", "int"),
        ("s_division_name", "string"),
        ("s_company_id", "int"),
        ("s_company_name", "string"),
        ("s_street_number", "string"),
        ("s_street_name", "string"),
        ("s_street_type", "string"),
        ("s_suite_number", "string"),
        ("s_city", "string"),
        ("s_county", "string"),
        ("s_state", "string"),
        ("s_zip", "string"),
        ("s_country", "string"),
        ("s_gmt_offset", "decimal(5,2)"),
        ("s_tax_percentage", "decimal(5,2)"),
    ],
    "tpcds_store_sales": [
        ("ss_sold_time_sk", "bigint"),
        ("ss_item_sk", "bigint"),
        ("ss_customer_sk", "bigint"),
        ("ss_cdemo_sk", "bigint"),
        ("ss_hdemo_sk", "bigint"),
        ("ss_addr_sk", "bigint"),
        ("ss_store_sk", "bigint"),
        ("ss_promo_sk", "bigint"),
        ("ss_ticket_number", "int"),
        ("ss_quantity", "int"),
        ("ss_wholesale_cost", "decimal(7,2)"),
        ("ss_list_price", "decimal(7,2)"),
        ("ss_sales_price", "decimal(7,2)"),
        ("ss_ext_discount_amt", "decimal(7,2)"),
        ("ss_ext_sales_price", "decimal(7,2)"),
        ("ss_ext_wholesale_cost", "decimal(7,2)"),
        ("ss_ext_list_price", "decimal(7,2)"),
        ("ss_ext_tax", "decimal(7,2)"),
        ("ss_coupon_amt", "decimal(7,2)"),
        ("ss_net_paid", "decimal(7,2)"),
        ("ss_net_paid_inc_tax", "decimal(7,2)"),
        ("ss_net_profit", "decimal(7,2)"),
        ("ss_sold_date_sk", "bigint"),
    ],
    "tpcds_time_dim": [
        ("t_time_sk", "bigint"),
        ("t_time_id", "string"),
        ("t_time", "int"),
        ("t_hour", "int"),
        ("t_minute", "int"),
        ("t_second", "int"),
        ("t_am_pm", "string"),
        ("t_shift", "string"),
        ("t_sub_shift", "string"),
        ("t_meal_time", "string"),
    ],
}


def table(name):
    return ibis.table(name=name, schema=MOCK_TABLES[name])


@pytest.fixture
def alltypes():
    return table("alltypes")



# Copyright 2014 Cloudera Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import pytest

import ibis
from ibis import config


@pytest.fixture
def queries(monkeypatch):
    queries = []
    monkeypatch.setattr(ibis.options, "verbose", True)
    monkeypatch.setattr(ibis.options, "verbose_log", queries.append)
    monkeypatch.setattr(ibis.options, "interactive", True)
    return queries


@pytest.fixture(scope="module")
def table(backend):
    return backend.functional_alltypes


@pytest.mark.notimpl(["dask", "pandas", "polars"])
def test_interactive_execute_on_repr(table, queries, snapshot):
    repr(table.bigint_col.sum())
    snapshot.assert_match(queries[0], "out.sql")


def test_repr_png_is_none_in_interactive(table):
    with config.option_context("interactive", True):
        assert table._repr_png_() is None


def test_repr_png_is_not_none_in_not_interactive(table):
    pytest.importorskip("ibis.expr.visualize")

    with config.option_context("interactive", False), config.option_context(
        "graphviz_repr", True
    ):
        assert table._repr_png_() is not None


@pytest.mark.notimpl(["dask", "pandas", "polars"])
def test_default_limit(table, snapshot, queries):
    repr(table.select("id", "bool_col"))

    snapshot.assert_match(queries[0], "out.sql")


@pytest.mark.notimpl(["dask", "pandas", "polars"])
def test_respect_set_limit(table, snapshot, queries):
    repr(table.select("id", "bool_col").limit(10))

    snapshot.assert_match(queries[0], "out.sql")


@pytest.mark.notimpl(["dask", "pandas", "polars"])
def test_disable_query_limit(table, snapshot, queries):
    assert ibis.options.sql.default_limit is None

    with config.option_context("sql.default_limit", 10):
        assert ibis.options.sql.default_limit == 10
        repr(table.select("id", "bool_col"))

    snapshot.assert_match(queries[0], "out.sql")


def test_interactive_non_compilable_repr_does_not_fail(table):
    """https://github.com/ibis-project/ibis/issues/170"""
    repr(table.string_col.topk(3))


def test_histogram_repr_no_query_execute(table, queries):
    tier = table.double_col.histogram(10).name("bucket")
    expr = table.group_by(tier).size()
    expr._repr()

    assert not queries


def test_isin_rule_suppressed_exception_repr_not_fail(table):
    bool_clause = table["string_col"].notin(["1", "4", "7"])
    expr = table[bool_clause]["string_col"].value_counts()

    repr(expr)



from __future__ import annotations

import functools
import ipaddress
import operator
import uuid
from collections import OrderedDict
from datetime import date, datetime, time
from decimal import Decimal
from operator import attrgetter, methodcaller

import numpy as np
import pytest
import pytz
import toolz
from pytest import param

import ibis
import ibis.common.exceptions as com
import ibis.expr.datashape as ds
import ibis.expr.datatypes as dt
import ibis.expr.operations as ops
import ibis.expr.types as ir
from ibis import _, literal
from ibis.common.annotations import ValidationError
from ibis.common.collections import frozendict
from ibis.common.exceptions import IbisTypeError
from ibis.expr import api
from ibis.tests.util import assert_equal


def test_null():
    assert ibis.literal(None).equals(ibis.null())
    assert ibis.null().op() == ops.Literal(None, dtype=dt.null)


@pytest.mark.parametrize(
    ["value", "expected_type"],
    [
        (5, "int8"),
        (127, "int8"),
        (128, "int16"),
        (32767, "int16"),
        (32768, "int32"),
        (2147483647, "int32"),
        (2147483648, "int64"),
        (-5, "int8"),
        (-128, "int8"),
        (-129, "int16"),
        (-32769, "int32"),
        (-2147483649, "int64"),
        (1.5, "float64"),
        ("foo", "string"),
        (b"fooblob", "bytes"),
        ([1, 2, 3], "array<int8>"),
        (ipaddress.ip_address("1.2.3.4"), "inet"),
        (ipaddress.ip_address("::1"), "inet"),
    ],
)
def test_literal_with_implicit_type(value, expected_type):
    expr = ibis.literal(value)

    assert isinstance(expr, ir.Scalar)
    assert expr.type() == dt.dtype(expected_type)

    assert isinstance(expr.op(), ops.Literal)
    assert expr.op().value == dt.normalize(expr.type(), value)


@pytest.mark.parametrize(
    ["value", "expected_type", "expected_value"],
    [
        ([1, 2, 3], "array<int8>", (1, 2, 3)),
        ([[1, 2], [3, 4]], "array<array<int8>>", ((1, 2), (3, 4))),
    ],
)
def test_listeral_with_unhashable_values(value, expected_type, expected_value):
    expr = ibis.literal(value)

    assert isinstance(expr, ir.Scalar)
    assert expr.type() == dt.dtype(expected_type)

    assert isinstance(expr.op(), ops.Literal)
    assert expr.op().value == expected_value


pointA = (1, 2)
pointB = (-3, 4)
pointC = (5, 19)
lineAB = [pointA, pointB]
lineBC = [pointB, pointC]
lineCA = [pointC, pointA]
polygon1 = [lineAB, lineBC, lineCA]
polygon2 = [lineAB, lineBC, lineCA]
multilinestring = [lineAB, lineBC, lineCA]
multipoint = [pointA, pointB, pointC]
multipolygon1 = [polygon1, polygon2]


@pytest.mark.parametrize(
    ["value", "expected_type"],
    [
        (5, "int16"),
        (127, "float64"),
        (128, "int64"),
        (32767, "float64"),
        (32768, "float32"),
        (2147483647, "int64"),
        (-5, "int16"),
        (-128, "int32"),
        (-129, "int64"),
        (-32769, "float32"),
        (-2147483649, "float64"),
        (1.5, "float64"),
        ("foo", "string"),
        (ipaddress.ip_address("1.2.3.4"), "inet"),
        (ipaddress.ip_address("::1"), "inet"),
    ],
)
def test_literal_with_explicit_type(value, expected_type):
    expr = ibis.literal(value, type=expected_type)
    assert expr.type().equals(dt.validate_type(expected_type))


@pytest.mark.parametrize(
    ["value", "expected_type"],
    [
        (list(pointA), "point"),
        (tuple(pointA), "point"),
        (list(lineAB), "linestring"),
        (tuple(lineAB), "linestring"),
        (list(polygon1), "polygon"),
        (tuple(polygon1), "polygon"),
        (list(multilinestring), "multilinestring"),
        (tuple(multilinestring), "multilinestring"),
        (list(multipoint), "multipoint"),
        (tuple(multipoint), "multipoint"),
        (list(multipolygon1), "multipolygon"),
        (tuple(multipolygon1), "multipolygon"),
        param(uuid.uuid4(), "uuid", id="uuid"),
        param(str(uuid.uuid4()), "uuid", id="uuid_str"),
        param(Decimal("234.234"), "decimal(6, 3)", id="decimal_native"),
        param(234234, "decimal(9, 3)", id="decimal_int"),
    ],
)
def test_literal_with_explicit_geotype(value, expected_type):
    pytest.importorskip("shapely")
    expr = ibis.literal(value, type=expected_type)
    assert expr.type().equals(dt.validate_type(expected_type))


@pytest.mark.parametrize(
    ("value", "expected", "dtype"),
    [
        # precision > scale
        (Decimal("234.234"), Decimal("234.234"), "decimal(6, 3)"),
        (234234, Decimal("234234.000"), "decimal(9, 3)"),
        # scale == 0
        (Decimal("234"), Decimal("234"), "decimal(6, 0)"),
        (234, Decimal("234"), "decimal(6, 0)"),
        # precision == scale
        (Decimal(".234"), Decimal(".234"), "decimal(3, 3)"),
        (234, Decimal("234.000"), "decimal(6, 3)"),
    ],
)
def test_normalize_decimal_literal(value, expected, dtype):
    expr = ibis.literal(value, type=dtype)
    assert expr.op().value == expected


@pytest.mark.parametrize(
    ["value", "expected_type", "expected_class"],
    [
        (list("abc"), "array<string>", ir.ArrayScalar),
        ([1, 2, 3], "array<int8>", ir.ArrayScalar),
        ({"a": 1, "b": 2, "c": 3}, "map<string, int8>", ir.MapScalar),
        ({1: 2, 3: 4, 5: 6}, "map<int8, int8>", ir.MapScalar),
        (
            {"a": [1.0, 2.0], "b": [], "c": [3.0]},
            "map<string, array<double>>",
            ir.MapScalar,
        ),
        (
            OrderedDict(
                [
                    ("a", 1),
                    ("b", list("abc")),
                    ("c", OrderedDict([("foo", [1.0, 2.0])])),
                ]
            ),
            "struct<a: int8, b: array<string>, c: struct<foo: array<double>>>",
            ir.StructScalar,
        ),
    ],
)
def test_literal_complex_types(value, expected_type, expected_class):
    expr = ibis.literal(value)
    expr_type = expr.type()
    assert expr_type.equals(dt.validate_type(expected_type))
    assert isinstance(expr, expected_class)
    assert isinstance(expr.op(), ops.Literal)
    assert expr.op().value == dt.normalize(dt.dtype(expected_type), value)


def test_simple_map_operations():
    value = {"a": [1.0, 2.0], "b": [], "c": [3.0]}
    value2 = {"a": [1.0, 2.0], "c": [3.0], "d": [4.0, 5.0]}
    expr = ibis.literal(value)
    expr2 = ibis.literal(value2)
    assert isinstance(expr, ir.MapValue)
    assert isinstance(expr.length().op(), ops.MapLength)
    assert isinstance((expr + expr2).op(), ops.MapMerge)
    assert isinstance((expr2 + expr).op(), ops.MapMerge)

    default = ibis.literal([0.0])
    assert isinstance(expr.get("d", default).op(), ops.MapGet)

    # test for an invalid default type, nulls are ok
    with pytest.raises(IbisTypeError):
        expr.get("d", ibis.literal("foo"))

    assert isinstance(expr.get("d", ibis.null()).op(), ops.MapGet)

    assert isinstance(expr["b"].op(), ops.MapGet)
    assert isinstance(expr.keys().op(), ops.MapKeys)
    assert isinstance(expr.values().op(), ops.MapValues)


@pytest.mark.parametrize(
    ["value", "expected_type"],
    [
        (32767, "int8"),
        (32768, "int16"),
        (2147483647, "int16"),
        (2147483648, "int32"),
    ],
)
def test_literal_with_non_coercible_type(value, expected_type):
    with pytest.raises(TypeError, match="out of bounds"):
        ibis.literal(value, type=expected_type)


def test_literal_double_from_string_fails():
    with pytest.raises(TypeError, match="Unable to normalize"):
        ibis.literal("foo", type="double")


def test_list_and_tuple_literals():
    what = [1, 2, 1000]
    expr = api.literal(what)
    assert isinstance(expr, ir.ArrayScalar)
    # it works!
    repr(expr)

    what = (1, 2, 1000)
    expr = api.literal(what)
    assert isinstance(expr, ir.ArrayScalar)
    # it works!
    repr(expr)


def test_list_and_tuple_literals_geotype():
    pytest.importorskip("shapely")
    # test using explicit type
    point = ibis.literal((1, 2, 1000), type="point")
    assert point.type() == dt.point
    point = ibis.literal([1, 2, 1000], type="point")
    assert point.type() == dt.point


def test_literal_array():
    what = []
    expr = api.literal(what)
    assert isinstance(expr, ir.ArrayValue)
    assert expr.type().equals(dt.Array(dt.null))


@pytest.mark.parametrize("container", [list, tuple, set, frozenset])
def test_isin_notin_list(table, container):
    values = container([1, 2, 3, 4])

    expr = table.a.isin(values)
    not_expr = table.a.notin(values)

    assert isinstance(expr, ir.BooleanColumn)
    assert isinstance(expr.op(), ops.InValues)

    assert isinstance(not_expr, ir.BooleanColumn)
    assert isinstance(not_expr.op(), ops.Not)
    assert isinstance(not_expr.op().arg, ops.InValues)


def test_value_counts(table, string_col):
    bool_clause = table[string_col].notin(["1", "4", "7"])
    expr = table[bool_clause][string_col].value_counts()
    assert isinstance(expr, ir.Table)


def test_isin_notin_scalars():
    a, b, c = (ibis.literal(x) for x in [1, 1, 2])

    result = a.isin([1, 2])
    assert isinstance(result, ir.BooleanScalar)

    result = a.notin([b, c, 3])
    assert isinstance(result, ir.BooleanScalar)


def test_scalar_isin_list_with_array(table):
    val = ibis.literal(2)

    options = [table.a, table.b, table.c]

    expr = val.isin(options)
    assert isinstance(expr, ir.BooleanColumn)

    not_expr = val.notin(options)
    assert isinstance(not_expr, ir.BooleanColumn)


def test_distinct_table(functional_alltypes):
    expr = functional_alltypes.distinct()
    assert isinstance(expr.op(), ops.Distinct)
    assert isinstance(expr, ir.Table)
    assert expr.op().parent == functional_alltypes.op()


def test_nunique(functional_alltypes):
    expr = functional_alltypes.string_col.nunique()
    assert isinstance(expr.op(), ops.CountDistinct)


def test_isnull(table):
    expr = table["g"].isnull()
    assert isinstance(expr, ir.BooleanColumn)
    assert isinstance(expr.op(), ops.IsNull)

    expr = ibis.literal("foo").isnull()
    assert isinstance(expr, ir.BooleanScalar)
    assert isinstance(expr.op(), ops.IsNull)


def test_notnull(table):
    expr = table["g"].notnull()
    assert isinstance(expr, ir.BooleanColumn)
    assert isinstance(expr.op(), ops.NotNull)

    expr = ibis.literal("foo").notnull()
    assert isinstance(expr, ir.BooleanScalar)
    assert isinstance(expr.op(), ops.NotNull)


@pytest.mark.parametrize("column", ["e", "f"], ids=["float32", "double"])
def test_isnan_isinf_column(table, column):
    expr = table[column].isnan()
    assert isinstance(expr, ir.BooleanColumn)
    assert isinstance(expr.op(), ops.IsNan)

    expr = table[column].isinf()
    assert isinstance(expr, ir.BooleanColumn)
    assert isinstance(expr.op(), ops.IsInf)


@pytest.mark.parametrize("value", [1.3, np.nan, np.inf, -np.inf])
def test_isnan_isinf_scalar(value):
    expr = ibis.literal(value).isnan()
    assert isinstance(expr, ir.BooleanScalar)
    assert isinstance(expr.op(), ops.IsNan)

    expr = ibis.literal(value).isinf()
    assert isinstance(expr, ir.BooleanScalar)
    assert isinstance(expr.op(), ops.IsInf)


@pytest.mark.parametrize(
    ["column", "operation"],
    [
        ("d", "cumsum"),
        ("d", "cummean"),
        ("d", "cummin"),
        ("d", "cummax"),
        ("h", "cumany"),
        ("h", "cumall"),
    ],
)
def test_cumulative_yield_array_types(table, column, operation):
    expr = getattr(getattr(table, column), operation)()
    assert isinstance(expr, ir.Column)


@pytest.fixture(params=["ln", "log", "log2", "log10"])
def log(request):
    return operator.methodcaller(request.param)


@pytest.mark.parametrize("column", list("abcdef"))
def test_log(table, log, column):
    result = log(table[column])
    assert isinstance(result, ir.FloatingColumn)

    # is this what we want?
    # assert result.get_name() == c


def test_log_string(table):
    g = table.g

    with pytest.raises(ValidationError):
        ops.Log(g, None).to_expr()


@pytest.mark.parametrize("klass", [ops.Ln, ops.Log2, ops.Log10])
def test_log_variants_string(table, klass):
    g = table.g

    with pytest.raises(ValidationError):
        klass(g).to_expr()


def test_log_boolean(table, log):
    # boolean not implemented for these
    h = table["h"]
    with pytest.raises(ValidationError):
        log(h)


def test_log_literal(log):
    assert isinstance(log(ibis.literal(5)), ir.FloatingScalar)
    assert isinstance(log(ibis.literal(5.5)), ir.FloatingScalar)


def test_cast_same_type_noop(table):
    c = table.g
    assert c.cast("string") is c

    i = ibis.literal(5)
    assert i.cast("int8") is i


@pytest.mark.parametrize("type", ["int8", "int32", "double", "float32"])
def test_string_to_number(table, type):
    casted = table.g.cast(type)
    casted_literal = ibis.literal("5").cast(type).name("bar")

    assert isinstance(casted, ir.Column)
    assert casted.type() == dt.dtype(type)

    assert isinstance(casted_literal, ir.Scalar)
    assert casted_literal.type() == dt.dtype(type)
    assert casted_literal.get_name() == "bar"


@pytest.mark.parametrize("col", list("abcdefh"))
def test_number_to_string_column(table, col):
    casted = table[col].cast("string")
    assert isinstance(casted, ir.StringColumn)


def test_number_to_string_scalar():
    casted_literal = ibis.literal(5).cast("string").name("bar")
    assert isinstance(casted_literal, ir.StringScalar)
    assert casted_literal.get_name() == "bar"


def test_casted_exprs_are_named(table):
    expr = table.f.cast("string")
    assert expr.get_name() == "Cast(f, string)"

    # it works! per GH #396
    expr.value_counts()


@pytest.mark.parametrize("col", list("abcdef"))
def test_negate(table, col):
    c = table[col]
    result = -c
    assert isinstance(result, type(c))
    assert isinstance(result.op(), ops.Negate)


@pytest.mark.parametrize("op", [operator.neg, operator.invert])
@pytest.mark.parametrize("value", [True, False])
def test_negate_boolean_scalar(op, value):
    result = op(ibis.literal(value))
    assert isinstance(result, ir.BooleanScalar)
    assert isinstance(result.op(), ops.Not)


@pytest.mark.parametrize("op", [operator.neg, operator.invert])
def test_negate_boolean_column(table, op):
    result = op(table["h"])
    assert isinstance(result, ir.BooleanColumn)
    assert isinstance(result.op(), ops.Not)


@pytest.mark.parametrize("column", ["a", "b", "c", "d", "e", "f", "g", "h"])
@pytest.mark.parametrize("how", ["first", "last", "heavy"])
@pytest.mark.parametrize("condition_fn", [lambda t: None, lambda t: t.a > 8])
def test_arbitrary(table, column, how, condition_fn):
    col = table[column]
    where = condition_fn(table)
    expr = col.arbitrary(how=how, where=where)
    assert expr.type() == col.type()
    assert isinstance(expr, ir.Scalar)
    assert isinstance(expr.op(), ops.Arbitrary)


@pytest.mark.parametrize(
    ["column", "operation"],
    [
        ("h", lambda column: column.any()),
        ("h", lambda column: column.notany()),
        ("h", lambda column: column.all()),
        ("c", lambda column: (column == 0).any()),
        ("c", lambda column: (column == 0).all()),
    ],
)
def test_any_all_notany(table, column, operation):
    expr = operation(table[column])
    assert isinstance(expr, ir.BooleanScalar)
    assert expr.op().find(ops.Reduction)


@pytest.mark.parametrize(
    "operation",
    [
        operator.lt,
        operator.gt,
        operator.ge,
        operator.le,
        operator.eq,
        operator.ne,
    ],
)
@pytest.mark.parametrize("column", list("abcdef"))
@pytest.mark.parametrize("case", [2, 2**9, 2**17, 2**33, 1.5])
def test_numbers_compare_numeric_literal(table, operation, column, case):
    ex_op_class = {
        operator.eq: ops.Equals,
        operator.ne: ops.NotEquals,
        operator.le: ops.LessEqual,
        operator.lt: ops.Less,
        operator.ge: ops.GreaterEqual,
        operator.gt: ops.Greater,
    }

    col = table[column]

    result = operation(col, case)
    assert isinstance(result, ir.BooleanColumn)
    assert isinstance(result.op(), ex_op_class[operation])


def test_boolean_comparisons(table):
    bool_col = table.h

    result = bool_col == True  # noqa: E712
    assert isinstance(result, ir.BooleanColumn)

    result = bool_col == False  # noqa: E712
    assert isinstance(result, ir.BooleanColumn)


@pytest.mark.parametrize(
    "operation",
    [
        operator.lt,
        operator.gt,
        operator.ge,
        operator.le,
        operator.eq,
        operator.ne,
    ],
)
def test_string_comparisons(table, operation):
    string_col = table.g
    result = operation(string_col, "foo")
    assert isinstance(result, ir.BooleanColumn)


@pytest.mark.parametrize("operation", [operator.xor, operator.or_, operator.and_])
def test_boolean_logical_ops(table, operation):
    expr = table.a > 0

    result = operation(expr, table.h)
    assert isinstance(result, ir.BooleanColumn)

    result = operation(expr, True)
    refl_result = operation(True, expr)
    assert isinstance(result, ir.BooleanColumn)
    assert isinstance(refl_result, ir.BooleanColumn)

    true = ibis.literal(True)
    false = ibis.literal(False)

    result = operation(true, false)
    assert isinstance(result, ir.BooleanScalar)


def test_and_(table):
    p1 = table.a > 1
    p2 = table.b > 1
    p3 = table.c > 1
    assert ibis.and_().equals(ibis.literal(True))
    assert ibis.and_(p1).equals(p1)
    assert ibis.and_(p1, p2).equals(p1 & p2)
    assert ibis.and_(p1, p2, p3).equals(p1 & p2 & p3)


def test_or_(table):
    p1 = table.a > 1
    p2 = table.b > 1
    p3 = table.c > 1
    assert ibis.or_().equals(ibis.literal(False))
    assert ibis.or_(p1).equals(p1)
    assert ibis.or_(p1, p2).equals(p1 | p2)
    assert ibis.or_(p1, p2, p3).equals(p1 | p2 | p3)


def test_null_column():
    t = ibis.table([("a", "string")], name="t")
    s = t.mutate(b=ibis.NA)
    assert s.b.type() == dt.null
    assert isinstance(s.b, ir.NullColumn)


def test_null_column_union():
    s = ibis.table([("a", "string"), ("b", "double")])
    t = ibis.table([("a", "string")])
    with pytest.raises(ibis.common.exceptions.RelationError):
        s.union(t.mutate(b=ibis.NA))  # needs a type
    assert s.union(t.mutate(b=ibis.NA.cast("double"))).schema() == s.schema()


def test_string_compare_numeric_array(table):
    with pytest.raises(com.IbisTypeError):
        table.g == table.f  # noqa: B015

    with pytest.raises(com.IbisTypeError):
        table.g == table.c  # noqa: B015


def test_string_compare_numeric_literal(table):
    with pytest.raises(com.IbisTypeError):
        table.g == ibis.literal(1.5)  # noqa: B015

    with pytest.raises(com.IbisTypeError):
        table.g == ibis.literal(5)  # noqa: B015


def test_between(table):
    result = table.f.between(0, 1)

    assert isinstance(result, ir.BooleanColumn)
    assert isinstance(result.op(), ops.Between)

    # it works!
    result = table.g.between("a", "f")
    assert isinstance(result, ir.BooleanColumn)

    result = ibis.literal(1).between(table.a, table.c)
    assert isinstance(result, ir.BooleanColumn)

    result = ibis.literal(7).between(5, 10)
    assert isinstance(result, ir.BooleanScalar)

    # Cases where between should immediately fail, e.g. incomparables
    with pytest.raises(ValidationError):
        table.f.between("0", "1")

    with pytest.raises(ValidationError):
        table.f.between(0, "1")

    with pytest.raises(ValidationError):
        table.f.between("0", 1)


def test_chained_comparisons_not_allowed(table):
    with pytest.raises(ValueError):
        0 < table.f < 1  # noqa: B015


@pytest.mark.parametrize(
    "operation",
    [operator.add, operator.sub, operator.truediv],
)
@pytest.mark.parametrize(("left", "right"), [("d", "g"), ("g", "d")])
def test_binop_string_type_error(table, operation, left, right):
    a = table[left]
    b = table[right]

    with pytest.raises((TypeError, ValidationError)):
        operation(a, b)


@pytest.mark.parametrize(("left", "right"), [("d", "g"), ("g", "d")])
def test_string_mul(table, left, right):
    a = table[left]
    b = table[right]

    expr = a * b
    assert isinstance(expr, ir.StringColumn)
    assert isinstance(expr.op(), ops.Repeat)


@pytest.mark.parametrize(
    ["op", "name", "case", "ex_type"],
    [
        (operator.add, "a", 0, "int8"),
        (operator.add, "a", 5, "int16"),
        (operator.add, "a", 100000, "int32"),
        (operator.add, "a", -100000, "int32"),
        (operator.add, "a", 1.5, "double"),
        (operator.add, "b", 0, "int16"),
        (operator.add, "b", 5, "int32"),
        (operator.add, "b", -5, "int32"),
        (operator.add, "c", 0, "int32"),
        (operator.add, "c", 5, "int64"),
        (operator.add, "c", -5, "int64"),
        # technically this can overflow, but we allow it
        (operator.add, "d", 5, "int64"),
        (operator.mul, "a", 0, "int8"),
        (operator.mul, "a", 5, "int16"),
        (operator.mul, "a", 2**24, "int32"),
        (operator.mul, "a", -(2**24) + 1, "int32"),
        (operator.mul, "a", 1.5, "double"),
        (operator.mul, "b", 0, "int16"),
        (operator.mul, "b", 5, "int32"),
        (operator.mul, "b", -5, "int32"),
        (operator.mul, "c", 0, "int32"),
        (operator.mul, "c", 5, "int64"),
        (operator.mul, "c", -5, "int64"),
        # technically this can overflow, but we allow it
        (operator.mul, "d", 5, "int64"),
        (operator.sub, "a", 5, "int16"),
        (operator.sub, "a", 100000, "int32"),
        (operator.sub, "a", -100000, "int32"),
        (operator.sub, "a", 1.5, "double"),
        (operator.sub, "b", 5, "int32"),
        (operator.sub, "b", -5, "int32"),
        (operator.sub, "c", 5, "int64"),
        (operator.sub, "c", -5, "int64"),
        # technically this can overflow, but we allow it
        (operator.sub, "d", 5, "int64"),
        (operator.truediv, "a", 5, "double"),
        (operator.truediv, "a", 1.5, "double"),
        (operator.truediv, "b", 5, "double"),
        (operator.truediv, "b", -5, "double"),
        (operator.truediv, "c", 5, "double"),
        (operator.pow, "a", 0, "double"),
        (operator.pow, "b", 0, "double"),
        (operator.pow, "c", 0, "double"),
        (operator.pow, "d", 0, "double"),
        (operator.pow, "e", 0, "float32"),
        (operator.pow, "f", 0, "double"),
        (operator.pow, "a", 2, "double"),
        (operator.pow, "b", 2, "double"),
        (operator.pow, "c", 2, "double"),
        (operator.pow, "d", 2, "double"),
        (operator.pow, "a", 1.5, "double"),
        (operator.pow, "b", 1.5, "double"),
        (operator.pow, "c", 1.5, "double"),
        (operator.pow, "d", 1.5, "double"),
        (operator.pow, "e", 2, "float32"),
        (operator.pow, "f", 2, "double"),
        (operator.pow, "a", -2, "double"),
        (operator.pow, "b", -2, "double"),
        (operator.pow, "c", -2, "double"),
        (operator.pow, "d", -2, "double"),
    ],
    ids=lambda arg: str(getattr(arg, "__name__", arg)),
)
def test_literal_promotions(table, op, name, case, ex_type):
    col = table[name]

    result = op(col, case)
    assert result.type() == dt.dtype(ex_type)

    result = op(case, col)
    assert result.type() == dt.dtype(ex_type)


@pytest.mark.parametrize(
    ("op", "left_fn", "right_fn", "ex_type"),
    [
        (operator.sub, lambda t: t["a"], lambda t: 0, "int8"),
        (operator.sub, lambda t: 0, lambda t: t["a"], "int16"),
        (operator.sub, lambda t: t["b"], lambda t: 0, "int16"),
        (operator.sub, lambda t: 0, lambda t: t["b"], "int32"),
        (operator.sub, lambda t: t["c"], lambda t: 0, "int32"),
        (operator.sub, lambda t: 0, lambda t: t["c"], "int64"),
    ],
    ids=lambda arg: str(getattr(arg, "__name__", arg)),
)
def test_zero_subtract_literal_promotions(table, op, left_fn, right_fn, ex_type):
    # in case of zero subtract the order of operands matters
    left, right = left_fn(table), right_fn(table)
    result = op(left, right)

    assert result.type() == dt.dtype(ex_type)


def test_substitute_dict():
    table = ibis.table([("foo", "string"), ("bar", "string")], "t1")
    subs = {"a": "one", "b": table.bar}

    result = table.foo.substitute(subs)
    expected = (
        ibis.case()
        .when(table.foo == "a", "one")
        .when(table.foo == "b", table.bar)
        .else_(table.foo)
        .end()
    )
    assert_equal(result, expected)

    result = table.foo.substitute(subs, else_=ibis.NA)
    expected = (
        ibis.case()
        .when(table.foo == "a", "one")
        .when(table.foo == "b", table.bar)
        .else_(ibis.NA)
        .end()
    )
    assert_equal(result, expected)


@pytest.mark.parametrize(
    "typ",
    [
        "array<map<string, array<array<double>>>>",
        "string",
        "double",
        "float32",
    ],
)
def test_not_without_boolean(typ):
    t = ibis.table([("a", typ)], name="t")
    c = t.a
    with pytest.raises(TypeError):
        ~c  # noqa: B018


@pytest.mark.parametrize(
    ("fn", "expected_op"),
    [
        param(lambda t: t.int_col & t.smallint_col, ops.BitwiseAnd, id="and"),
        param(lambda t: t.int_col | t.smallint_col, ops.BitwiseOr, id="or"),
        param(lambda t: t.int_col ^ t.smallint_col, ops.BitwiseXor, id="xor"),
        param(
            lambda t: t.int_col << t.smallint_col,
            ops.BitwiseLeftShift,
            id="lshift",
        ),
        param(
            lambda t: t.int_col >> t.smallint_col,
            ops.BitwiseRightShift,
            id="rshift",
        ),
        param(lambda t: operator.inv(t.int_col), ops.BitwiseNot, id="not"),
    ],
)
def test_bitwise_exprs(fn, expected_op):
    t = ibis.table(dict(int_col="int32", smallint_col="int16"), name="t")
    expr = fn(t)
    assert isinstance(expr.op(), expected_op)


@pytest.mark.parametrize(
    ("position", "names"),
    [
        (0, "foo"),
        (1, "bar"),
        ([0], ["foo"]),
        ([1], ["bar"]),
        ([0, 1], ["foo", "bar"]),
        ([1, 0], ["bar", "foo"]),
    ],
)
@pytest.mark.parametrize(
    "expr_func",
    [
        lambda t, args: t[args],
        lambda t, args: t.order_by(args),
        lambda t, args: t.group_by(args).aggregate(bar_avg=t.bar.mean()),
    ],
)
def test_table_operations_with_integer_column(position, names, expr_func):
    t = ibis.table([("foo", "string"), ("bar", "double")])
    result = expr_func(t, position)
    expected = expr_func(t, names)
    assert result.equals(expected)


@pytest.mark.parametrize("value", ["abcdefg", ["a", "b", "c"], [1, 2, 3]])
@pytest.mark.parametrize("operation", ["pow", "sub", "truediv", "floordiv", "mod"])
def test_generic_value_api_no_arithmetic(value, operation):
    func = getattr(operator, operation)
    expr = ibis.literal(value)
    with pytest.raises(TypeError):
        func(expr, expr)


@pytest.mark.parametrize(
    ("value", "expected"), [(5, dt.int8), (5.4, dt.double), ("abc", dt.string)]
)
def test_fillna_null(value, expected):
    assert ibis.NA.fillna(value).type().equals(expected)


@pytest.mark.parametrize(
    ("left", "right"),
    [
        (literal("2017-04-01"), date(2017, 4, 2)),
        (date(2017, 4, 2), literal("2017-04-01")),
        (literal("2017-04-01 01:02:33"), datetime(2017, 4, 1, 1, 3, 34)),
        (datetime(2017, 4, 1, 1, 3, 34), literal("2017-04-01 01:02:33")),
    ],
)
@pytest.mark.parametrize(
    "op",
    [
        operator.eq,
        operator.ne,
        operator.lt,
        operator.le,
        operator.gt,
        operator.ge,
        lambda left, right: ibis.timestamp("2017-04-01 00:02:34").between(left, right),
        lambda left, right: ibis.timestamp("2017-04-01")
        .cast(dt.date)
        .between(left, right),
    ],
)
def test_string_temporal_compare(op, left, right):
    result = op(left, right)
    assert result.type().equals(dt.boolean)


@pytest.mark.parametrize(
    ("value", "type", "expected_type_class"),
    [
        (2.21, "decimal", dt.Decimal),
        (3.14, "float64", dt.Float64),
        (4.2, "int64", dt.Float64),
        (4, "int64", dt.Int64),
    ],
)
def test_decimal_modulo_output_type(value, type, expected_type_class):
    t = ibis.table([("a", type)])
    expr = t.a % value
    assert isinstance(expr.type(), expected_type_class)


@pytest.mark.parametrize(
    ("left", "right"),
    [(literal("10:00"), time(10, 0)), (time(10, 0), literal("10:00"))],
)
@pytest.mark.parametrize(
    "op",
    [
        operator.eq,
        operator.ne,
        operator.lt,
        operator.le,
        operator.gt,
        operator.ge,
    ],
)
def test_time_compare(op, left, right):
    result = op(left, right)
    assert result.type().equals(dt.boolean)


@pytest.mark.parametrize(
    ("left", "right"),
    [
        (literal("10:00"), date(2017, 4, 2)),
        (literal("10:00"), datetime(2017, 4, 2, 1, 1)),
        (literal("10:00"), literal("2017-04-01")),
    ],
)
@pytest.mark.parametrize(
    "op", [operator.eq, operator.lt, operator.le, operator.gt, operator.ge]
)
def test_time_timestamp_invalid_compare(op, left, right):
    result = op(left, right)
    assert result.type().equals(dt.boolean)


@pytest.mark.parametrize(
    ("left", "right", "expected"),
    [
        (
            # same value type, same name
            ibis.param(dt.timestamp),
            ibis.param(dt.timestamp),
            False,
        ),
        (
            # different value type, same name
            ibis.param(dt.date),
            ibis.param(dt.timestamp),
            False,
        ),
        (
            # same value type, different name
            ibis.param(dt.timestamp),
            ibis.param(dt.timestamp),
            False,
        ),
        (
            # different value type, different name
            ibis.param(dt.date),
            ibis.param(dt.timestamp),
            False,
        ),
    ],
)
def test_scalar_parameter_compare(left, right, expected):
    assert left.equals(right) == expected


@pytest.mark.parametrize(
    ("left", "right"),
    [
        # different Python class, left side is param
        param(ibis.param(dt.timestamp), dt.date, id="param_dtype"),
        # different Python class, right side is param
        param(dt.date, ibis.param(dt.timestamp), id="dtype_param"),
    ],
)
def test_scalar_parameter_invalid_compare(left, right):
    with pytest.raises(TypeError):
        left.equals(right)


@pytest.mark.parametrize(
    ("case", "creator"),
    [
        (datetime.now(), toolz.compose(methodcaller("time"), ibis.timestamp)),
        ("now", toolz.compose(methodcaller("time"), ibis.timestamp)),
        (datetime.now().time(), ibis.time),
        ("10:37", ibis.time),
    ],
)
@pytest.mark.parametrize(
    ("left", "right"), [(1, "a"), ("a", 1), (1.0, 2.0), (["a"], [1])]
)
def test_between_time_failure_time(case, creator, left, right):
    value = creator(case)
    with pytest.raises(ValidationError):
        value.between(left, right)


def test_empty_array_as_argument():
    class Foo(ir.Expr):
        pass

    class FooNode(ops.Node):
        value: ops.Value[dt.Array[dt.Int64], ds.Any]

        def to_expr(self):
            return Foo(self)

    node = FooNode([])
    assert node.value.value == ()
    assert node.value.dtype == dt.Array(dt.Int64)


def test_nullable_column_propagated():
    t = ibis.table(
        [
            ("a", dt.Int32(nullable=True)),
            ("b", dt.Int32(nullable=False)),
            ("c", dt.String(nullable=False)),
            ("d", dt.float64),  # nullable by default
            ("f", dt.Float64(nullable=False)),
        ]
    )

    assert t.a.type().nullable is True
    assert t.b.type().nullable is False
    assert t.c.type().nullable is False
    assert t.d.type().nullable is True
    assert t.f.type().nullable is False

    s = t.a + t.d
    assert s.type().nullable is True

    s = t.b + t.d
    assert s.type().nullable is True

    s = t.b + t.f
    assert s.type().nullable is False


@pytest.mark.parametrize(
    "base_expr",
    [
        ibis.table([("interval_col", dt.Interval(unit="D"))]).interval_col,
        ibis.interval(seconds=42),
    ],
)
def test_interval_negate(base_expr):
    expr = -base_expr
    expr2 = base_expr.negate()
    assert isinstance(expr.op(), ops.Negate)
    assert expr.equals(expr2)


def test_large_timestamp():
    expr = ibis.timestamp("4567-02-03")
    expected = datetime(year=4567, month=2, day=3)
    result = expr.op().value
    assert result == expected


def test_timestamp_with_timezone():
    expr = ibis.timestamp("2017-01-01", timezone=None)
    expected = datetime(2017, 1, 1, tzinfo=None)
    assert expr.op().value == expected

    expr = ibis.timestamp("2017-01-01", timezone="UTC")
    expected = datetime(2017, 1, 1, tzinfo=pytz.timezone("UTC"))
    assert expr.op().value == expected


@pytest.mark.parametrize("tz", [None, "UTC"])
def test_timestamp_timezone_type(tz):
    expr = ibis.timestamp("2017-01-01", timezone=tz)
    expected = dt.Timestamp(timezone=tz)
    assert expected == expr.op().dtype


def test_map_get_broadcast():
    t = ibis.table([("a", "string")], name="t")
    lookup_table = ibis.literal({"a": 1, "b": 2})

    expr = lookup_table.get(t.a)
    assert isinstance(expr, ir.IntegerColumn)


def test_map_getitem_broadcast():
    t = ibis.table([("a", "string")], name="t")
    lookup_table = ibis.literal({"a": 1, "b": 2})
    expr = lookup_table[t.a]
    assert isinstance(expr, ir.IntegerColumn)


def test_map_keys_output_type():
    mapping = ibis.literal({"a": 1, "b": 2})
    assert mapping.keys().type() == dt.Array(dt.string)


def test_map_values_output_type():
    mapping = ibis.literal({"a": 1, "b": 2})
    assert mapping.values().type() == dt.Array(dt.int8)


def test_scalar_isin_map_keys():
    mapping = ibis.literal({"a": 1, "b": 2})
    key = ibis.literal("a")
    expr = key.isin(mapping.keys())
    assert isinstance(expr, ir.BooleanScalar)


def test_column_isin_array():
    # scalar case
    t = ibis.table([("a", "string")], name="t")
    expr = t.a.isin(ibis.array(["a", "b"]))
    assert isinstance(expr, ir.BooleanColumn)
    assert isinstance(expr.op(), ops.ArrayContains)
    assert expr.op().shape.is_columnar()

    # columnar case
    t = ibis.table([("a", "string"), ("b", "array<string>")], name="t")
    expr = t.a.isin(t.b)
    assert isinstance(expr, ir.BooleanColumn)
    assert isinstance(expr.op(), ops.ArrayContains)
    assert expr.op().shape.is_columnar()


def test_column_isin_map_keys():
    t = ibis.table([("a", "string")], name="t")
    mapping = ibis.literal({"a": 1, "b": 2})
    expr = t.a.isin(mapping.keys())
    assert isinstance(expr, ir.BooleanColumn)
    assert isinstance(expr.op(), ops.ArrayContains)


def test_map_get_with_compatible_value_smaller():
    value = ibis.literal({"A": 1000, "B": 2000})
    expr = value.get("C", 3)
    assert value.type() == dt.Map(dt.string, dt.int16)
    assert expr.type() == dt.int16


def test_map_get_with_compatible_value_bigger():
    value = ibis.literal({"A": 1, "B": 2})
    expr = value.get("C", 3000)
    assert value.type() == dt.Map(dt.string, dt.int8)
    assert expr.type() == dt.int16


def test_map_get_with_incompatible_value_different_kind():
    value = ibis.literal({"A": 1000, "B": 2000})
    assert value.get("C", 3.0).type() == dt.float64


@pytest.mark.parametrize("null_value", [None, ibis.NA])
def test_map_get_with_null_on_not_nullable(null_value):
    map_type = dt.Map(dt.string, dt.Int16(nullable=False))
    value = ibis.literal({"A": 1000, "B": 2000}).cast(map_type)
    assert value.type() == map_type
    expr = value.get("C", null_value)
    assert expr.type() == dt.Int16(nullable=True)


@pytest.mark.parametrize("null_value", [None, ibis.NA])
def test_map_get_with_null_on_nullable(null_value):
    value = ibis.literal({"A": 1000, "B": None})
    result = value.get("C", null_value)
    assert result.type().nullable


@pytest.mark.parametrize("null_value", [None, ibis.NA])
def test_map_get_with_null_on_null_type_with_null(null_value):
    value = ibis.literal({"A": None, "B": None})
    result = value.get("C", null_value)
    assert result.type().nullable


def test_map_get_with_null_on_null_type_with_non_null():
    value = ibis.literal({"A": None, "B": None})
    assert value.get("C", 1).type() == dt.int8


def test_map_get_with_incompatible_value():
    value = ibis.literal({"A": 1000, "B": 2000})
    with pytest.raises(TypeError):
        value.get("C", ["A"])


@pytest.mark.parametrize(
    ("value", "expected_type"),
    [
        (datetime.now(), dt.timestamp),
        (datetime.now().date(), dt.date),
        (datetime.now().time(), dt.time),
    ],
)
def test_invalid_negate(value, expected_type):
    expr = ibis.literal(value)
    assert expr.type() == expected_type
    with pytest.raises(TypeError):
        -expr  # noqa: B018


@pytest.mark.parametrize(
    "type",
    [
        np.float16,
        np.float32,
        np.float64,
        np.int16,
        np.int32,
        np.int64,
        np.int64,
        np.int8,
        np.timedelta64,
        np.uint16,
        np.uint32,
        np.uint64,
        np.uint64,
        np.uint8,
        float,
        int,
    ],
)
def test_valid_negate(type):
    expr = ibis.literal(1)
    assert -expr is not None


@pytest.mark.parametrize(
    ("left", "right", "expected"),
    [
        (ibis.literal(1), ibis.literal(1.0), dt.float64),
        (ibis.literal("a"), ibis.literal("b"), dt.string),
        (ibis.literal(1.0), ibis.literal(1), dt.float64),
        (ibis.literal(1), ibis.literal(1), dt.int8),
        (ibis.literal(1), ibis.literal(1000), dt.int16),
        (ibis.literal(2**16), ibis.literal(2**17), dt.int32),
        (ibis.literal(2**50), ibis.literal(1000), dt.int64),
        (ibis.literal([1, 2]), ibis.literal([1, 2]), dt.Array(dt.int8)),
        (ibis.literal(["a"]), ibis.literal([]), dt.Array(dt.string)),
        (ibis.literal([]), ibis.literal(["a"]), dt.Array(dt.string)),
        (ibis.literal([]), ibis.literal([]), dt.Array(dt.null)),
    ],
)
def test_nullif_type(left, right, expected):
    assert left.nullif(right).type() == expected


@pytest.mark.parametrize(("left", "right"), [(ibis.literal(1), ibis.literal("a"))])
def test_nullif_fail(left, right):
    with pytest.raises(com.IbisTypeError):
        left.nullif(right)
    with pytest.raises(com.IbisTypeError):
        right.nullif(left)


@pytest.mark.parametrize(
    "join_method",
    [
        "left_join",
        "right_join",
        "inner_join",
        "outer_join",
        "asof_join",
    ],
)
def test_select_on_unambiguous_join(join_method):
    t = ibis.table([("a0", dt.int64), ("b1", dt.string)], name="t")
    s = ibis.table([("a1", dt.int64), ("b2", dt.string)], name="s")
    method = getattr(t, join_method)
    join = method(s, t.b1 == s.b2)
    expr1 = join["a0", "a1"]
    expr2 = join[["a0", "a1"]]
    expr3 = join.select(["a0", "a1"])
    assert expr1.equals(expr2)
    assert expr1.equals(expr3)


def test_chained_select_on_join():
    t = ibis.table([("a", dt.int64)], name="t")
    s = ibis.table([("a", dt.int64), ("b", dt.string)], name="s")
    join = t.join(s)[t.a, s.b]
    expr1 = join["a", "b"]
    expr2 = join.select(["a", "b"])
    assert expr1.equals(expr2)


def test_repr_list_of_lists():
    lit = ibis.literal([[1]])
    repr(lit)


def test_repr_list_of_lists_in_table():
    t = ibis.table([("a", "int64")], name="t")
    lit = ibis.literal([[1]])
    expr = t[t, lit.name("array_of_array")]
    repr(expr)


@pytest.mark.parametrize(
    ("expr", "expected_type"),
    [
        (ibis.coalesce(ibis.NA, 1), dt.int8),
        (ibis.coalesce(1, ibis.NA), dt.int8),
        (ibis.coalesce(ibis.NA, 1000), dt.int16),
        (ibis.coalesce(ibis.NA), dt.null),
        (ibis.coalesce(ibis.NA, ibis.NA), dt.null),
        (
            ibis.coalesce(ibis.NA, ibis.NA.cast("array<string>")),
            dt.Array(dt.string),
        ),
    ],
)
def test_coalesce_type_inference_with_nulls(expr, expected_type):
    assert expr.type() == expected_type


def test_literal_hash():
    expr = ibis.literal(1)
    op = expr.op()
    # hashing triggers computation and storage of the op's hash
    result1 = hash(op)
    assert op.__precomputed_hash__ == result1
    assert hash(op) == result1


@pytest.mark.parametrize(
    "op_name",
    [
        "add",
        "sub",
        "mul",
        "truediv",
        "floordiv",
        "pow",
        "mod",
        "eq",
        "ne",
        "lt",
        "le",
        "gt",
        "ge",
    ],
)
@pytest.mark.parametrize(
    ("left", "right"),
    [
        param(_.a, 2, id="a_int"),
        param(2, _.a, id="int_a"),
        param(_.a, _.b, id="a_b"),
    ],
)
def test_deferred(op_name, left, right):
    t = ibis.table(dict(a="int64"), name="t")

    op = getattr(operator, op_name)
    expr = (
        t.mutate(b=_.a)
        .mutate(expr=op(left, right))
        .group_by(key=_.b.cast("string"))
        .aggregate(avg=_.b.mean(where=_.a < 1), neg_sum=-_.b.sum())
        .mutate(c=_.avg <= 1.0, d=~(_["neg_sum"] > 2))
        .filter(_.c)
    )

    assert expr.schema() == ibis.schema(
        dict(
            key="string",
            avg="float64",
            neg_sum="int64",
            c="bool",
            d="bool",
        )
    )


@pytest.mark.parametrize("bool_op_name", ["or_", "and_", "xor"])
@pytest.mark.parametrize(
    ("bool_left", "bool_right"),
    [
        param(_.a < 1, True, id="a_true"),
        param(False, _.a > 1, id="false_a"),
        param(_.a == 2, _.b == 1, id="expr_expr"),
    ],
)
def test_deferred_bool(bool_op_name, bool_left, bool_right):
    t = ibis.table(dict(a="int64"), name="t")

    bool_op = getattr(operator, bool_op_name)
    expr = t.mutate(b=_.a).mutate(c=bool_op(bool_left, bool_right)).filter(_.c)

    assert expr.schema() == ibis.schema(dict(a="int64", b="int64", c="bool"))


@pytest.mark.parametrize(
    ("op_name", "expected_left", "expected_right"),
    [
        param("add", attrgetter("a"), lambda _: ibis.literal(2), id="add"),
        param("sub", lambda _: ibis.literal(2), attrgetter("a"), id="sub"),
        param("mul", attrgetter("a"), lambda _: ibis.literal(2), id="mul"),
        param(
            "truediv",
            lambda _: ibis.literal(2),
            attrgetter("a"),
            id="truediv",
        ),
        param(
            "floordiv",
            lambda _: ibis.literal(2),
            attrgetter("a"),
            id="floordiv",
        ),
        param("pow", lambda _: ibis.literal(2), attrgetter("a"), id="pow"),
        param("mod", lambda _: ibis.literal(2), attrgetter("a"), id="mod"),
    ],
)
def test_deferred_r_ops(op_name, expected_left, expected_right):
    t = ibis.table(dict(a="int64"), name="t")

    left = 2
    right = _.a

    op = getattr(operator, op_name)
    expr = t[op(left, right).name("b")]
    node = expr.op().values["b"]
    assert node.left.equals(expected_left(t).op())
    assert node.right.equals(expected_right(t).op())


@pytest.mark.parametrize(
    ("expr_fn", "expected_type"),
    [
        (lambda t: ibis.ifelse(t.a == 1, t.b, ibis.NA), dt.string),
        (lambda t: ibis.ifelse(t.a == 1, t.b, t.a.cast("string")), dt.string),
        (
            lambda t: ibis.ifelse(t.a == 1, t.b, t.a.cast("!string")),
            dt.string.copy(nullable=False),
        ),
        (lambda _: ibis.ifelse(True, ibis.NA, ibis.NA), dt.null),
        (lambda _: ibis.ifelse(False, ibis.NA, ibis.NA), dt.null),
    ],
)
def test_non_null_with_null_precedence(expr_fn, expected_type):
    t = ibis.table(dict(a="int64", b="!string"), name="t")
    expr = expr_fn(t)
    assert expr.type() == expected_type


def test_struct_names_types_fields():
    s = ibis.struct(dict(a=1, b="2", c=[[1.0], [], [None]]))
    assert s.names == ("a", "b", "c")
    assert s.types == (dt.int8, dt.string, dt.Array(dt.Array(dt.float64)))
    assert s.fields == dict(a=dt.int8, b=dt.string, c=dt.Array(dt.Array(dt.float64)))
    assert isinstance(s.fields, frozendict)


@pytest.mark.parametrize(
    ["arg", "typestr", "type"],
    [
        ([1, 2, 3], None, dt.Array(dt.int8)),
        ([1, 2, 3], "array<int16>", dt.Array(dt.int16)),
        ([1, 2, 3.0], None, dt.Array(dt.double)),
        (["a", "b", "c"], None, dt.Array(dt.string)),
    ],
)
def test_array_literal(arg, typestr, type):
    x = ibis.literal(arg, type=typestr)
    assert x.op().value == tuple(arg)
    assert x.type() == type


def test_array_length_scalar():
    raw_value = [1, 2, 4]
    value = ibis.literal(raw_value)
    expr = value.length()
    assert isinstance(expr.op(), ops.ArrayLength)


def double_int(x):
    return x * 2


def double_float(x):
    return x * 2.0


def is_negative(x):
    return x < 0


def test_array_map():
    arr = ibis.array([1, 2, 3])

    result_int = arr.map(double_int)
    result_float = arr.map(double_float)

    assert result_int.type() == dt.Array(dt.int16)
    assert result_float.type() == dt.Array(dt.float64)


def test_array_map_partial():
    arr = ibis.array([1, 2, 3])

    def add(x, y):
        return x + y

    result = arr.map(functools.partial(add, y=2))
    assert result.type() == dt.Array(dt.int16)


def test_array_filter():
    arr = ibis.array([1, 2, 3])
    result = arr.filter(is_negative)
    assert result.type() == arr.type()


def test_array_filter_partial():
    arr = ibis.array([1, 2, 3])

    def equal(x, y):
        return x == y

    result = arr.filter(functools.partial(equal, y=2))
    assert result.type() == arr.type()


@pytest.mark.parametrize(
    ("func", "expected_type"),
    [
        param(ibis.timestamp, dt.timestamp, id="timestamp"),
        param(ibis.date, dt.date, id="date"),
        param(ibis.time, dt.time, id="time"),
        param(ibis.coalesce, dt.timestamp, id="coalesce"),
        param(ibis.greatest, dt.timestamp, id="greatest"),
        param(ibis.least, dt.timestamp, id="least"),
        param(
            lambda ts: ibis.ifelse(ts.notnull(), ts, ts - ibis.interval(days=1)),
            dt.timestamp,
            id="ifelse",
        ),
    ],
)
def test_deferred_function_call(func, expected_type):
    t = ibis.table(dict(ts="timestamp"), name="t")
    expr = t.select(col=func(_.ts))
    assert expr["col"].type() == expected_type


@pytest.mark.parametrize(
    "case",
    [
        param(lambda: (ibis.array([1, _]), ibis.array([1, 2])), id="array"),
        param(
            lambda: (ibis.map({"x": 1, "y": _}), ibis.map({"x": 1, "y": 2})), id="map"
        ),
        param(
            lambda: (ibis.struct({"x": 1, "y": _}), ibis.struct({"x": 1, "y": 2})),
            id="struct",
        ),
    ],
)
def test_deferred_nested_types(case):
    expr, sol = case()
    assert expr.resolve(2).equals(sol)


def test_numpy_ufuncs_dont_cast_columns():
    t = ibis.table(dict.fromkeys("abcd", "int"))

    # Adding a numpy array doesn't implicitly compute
    arr = np.array([1, 2, 3])
    for left, right in [(t.a, arr), (arr, t.a)]:
        with pytest.raises(TypeError):
            left + right

    # Adding a numpy scalar works and results in a new expr
    x = np.int64(1)
    for expr in [t.a + x, x + t.a]:
        assert expr.equals(t.a + ibis.literal(x))


@pytest.mark.parametrize(
    "operation",
    [operator.lt, operator.gt, operator.ge, operator.le, operator.eq, operator.ne],
)
def test_logical_comparison_rlz_incompatible_error(table, operation):
    with pytest.raises(com.IbisTypeError, match=r"b:int16 and Literal\(foo\):string"):
        operation(table.b, "foo")


def test_case_rlz_incompatible_error(table):
    with pytest.raises(com.IbisTypeError, match=r"a:int8 and Literal\(foo\):string"):
        table.a == "foo"  # noqa: B015


@pytest.mark.parametrize("func", [ibis.asc, ibis.desc])
def test_group_by_order_by_deferred(func):
    from ibis import _

    table = ibis.table(dict(x="string", y="int"), name="t")
    expr = table.group_by(_.x).aggregate(mean_y=_.y.mean()).order_by(func(_.mean_y))
    assert isinstance(expr, ir.Table)


def test_rowid_only_physical_tables():
    table = ibis.table({"x": "int", "y": "string"}, name="t")

    table.rowid()  # works
    table[table.rowid(), table.x].filter(_.x > 10)  # works
    with pytest.raises(com.IbisTypeError, match="only valid for physical tables"):
        table.filter(table.x > 0).rowid()


def test_where_shape():
    # GH-5191
    t = ibis.table(dict(a="int64", b="string"), name="t")
    expr = ibis.literal(True).ifelse(t.a, -t.a)
    assert isinstance(expr, ir.IntegerColumn)
    assert isinstance(expr, ir.ColumnExpr)


def test_quantile_shape():
    t = ibis.table([("a", "float64")])

    b1 = t.a.quantile(0.25).name("br2")
    assert isinstance(b1, ir.Scalar)

    projs = [b1]
    expr = t.select(projs)
    b1 = expr.br2

    assert b1.op().shape.is_columnar()


def test_sample():
    t = ibis.table({"x": "int64", "y": "string"})

    expr = t.sample(1)
    assert expr.equals(t)

    expr = t.sample(0)
    assert expr.equals(t.limit(0))

    expr = t.sample(0.5, method="block", seed=1234)
    assert expr.schema() == t.schema()
    op = expr.op()
    assert op.fraction == 0.5
    assert op.method == "block"
    assert op.seed == 1234


def test_deferred_doesnt_convert_callables():
    t = ibis.table([("a", "int64"), ("b", "string")])
    expr = t.mutate(b=_.b.split(",").filter(lambda pp: ~pp.isin(("word1", "word2"))))
    expected = t.mutate(
        b=t.b.split(",").filter(lambda pp: ~pp.isin(("word1", "word2")))
    )
    assert expr.equals(expected)


def test_in_subquery_shape():
    t = ibis.table([("a", "int64"), ("b", "string")])

    expr = t.a.cast("string").isin(t.b)
    assert expr.op().shape.is_columnar()

    expr = ibis.literal(2).isin(t.a)
    assert expr.op().shape.is_scalar()



from __future__ import annotations

import concurrent.futures
import contextlib
import io
import os
from typing import Any

import google.api_core.exceptions as gexc
import google.auth
import pytest
from google.cloud import bigquery as bq

import ibis
from ibis.backends.bigquery import EXTERNAL_DATA_SCOPES, Backend
from ibis.backends.bigquery.datatypes import BigQuerySchema
from ibis.backends.conftest import TEST_TABLES
from ibis.backends.tests.base import BackendTest
from ibis.backends.tests.data import json_types, non_null_array_types, struct_types, win

DATASET_ID = "ibis_gbq_testing"
DATASET_ID_TOKYO = "ibis_gbq_testing_tokyo"
REGION_TOKYO = "asia-northeast1"
DEFAULT_PROJECT_ID = "ibis-gbq"
PROJECT_ID_ENV_VAR = "GOOGLE_BIGQUERY_PROJECT_ID"


class TestConf(BackendTest):
    """Backend-specific class with information for testing."""

    supports_divide_by_zero = True
    supports_floating_modulus = False
    returned_timestamp_unit = "us"
    supports_structs = True
    supports_json = True
    check_names = False
    force_sort = True
    deps = ("google.cloud.bigquery",)

    @staticmethod
    def format_table(name: str) -> str:
        return f"{DATASET_ID}.{name}"

    def _load_data(self, **_: Any) -> None:
        """Load test data into a BigQuery instance."""

        credentials, default_project_id = google.auth.default(
            scopes=EXTERNAL_DATA_SCOPES
        )

        project_id = (
            os.environ.get(PROJECT_ID_ENV_VAR, default_project_id) or DEFAULT_PROJECT_ID
        )

        client = bq.Client(project=project_id, credentials=credentials)

        try:
            client.query("SELECT 1")
        except gexc.Forbidden:
            pytest.skip("User does not have permission to create dataset")

        testing_dataset = bq.DatasetReference(project_id, DATASET_ID)

        with contextlib.suppress(gexc.NotFound):
            client.create_dataset(testing_dataset, exists_ok=True)

        testing_dataset_tokyo = bq.Dataset(
            bq.DatasetReference(project_id, DATASET_ID_TOKYO)
        )
        testing_dataset_tokyo.location = REGION_TOKYO

        with contextlib.suppress(gexc.NotFound):
            client.create_dataset(testing_dataset_tokyo, exists_ok=True)

        # day partitioning
        functional_alltypes_parted = bq.Table(
            bq.TableReference(testing_dataset, "functional_alltypes_parted")
        )
        functional_alltypes_parted.require_partition_filter = False
        functional_alltypes_parted.time_partitioning = bq.TimePartitioning(
            type_=bq.TimePartitioningType.DAY
        )

        # ingestion timestamp partitioning
        timestamp_table = bq.Table(
            bq.TableReference(testing_dataset, "timestamp_column_parted")
        )
        timestamp_table.schema = BigQuerySchema.from_ibis(
            ibis.schema(
                dict(
                    my_timestamp_parted_col="timestamp",
                    string_col="string",
                    int_col="int",
                )
            )
        )
        timestamp_table.time_partitioning = bq.TimePartitioning(
            field="my_timestamp_parted_col"
        )
        client.create_table(timestamp_table, exists_ok=True)

        # ingestion date partitioning
        date_table = bq.Table(bq.TableReference(testing_dataset, "date_column_parted"))
        date_table.schema = BigQuerySchema.from_ibis(
            ibis.schema(
                dict(my_date_parted_col="date", string_col="string", int_col="int")
            )
        )
        date_table.time_partitioning = bq.TimePartitioning(field="my_date_parted_col")
        client.create_table(date_table, exists_ok=True)

        write_disposition = bq.WriteDisposition.WRITE_TRUNCATE
        make_job = lambda func, *a, **kw: func(*a, **kw).result()

        futures = []
        # 10 is because of urllib3 connection pool size
        with concurrent.futures.ThreadPoolExecutor(max_workers=10) as e:
            futures.append(
                e.submit(
                    make_job,
                    client.load_table_from_dataframe,
                    struct_types,
                    bq.TableReference(testing_dataset, "struct"),
                    job_config=bq.LoadJobConfig(
                        write_disposition=write_disposition,
                        schema=BigQuerySchema.from_ibis(
                            ibis.schema(
                                dict(abc="struct<a: float64, b: string, c: int64>")
                            )
                        ),
                    ),
                )
            )

            futures.append(
                e.submit(
                    make_job,
                    client.load_table_from_dataframe,
                    non_null_array_types.drop(columns=["multi_dim"]),
                    bq.TableReference(testing_dataset, "array_types"),
                    job_config=bq.LoadJobConfig(
                        write_disposition=write_disposition,
                        schema=BigQuerySchema.from_ibis(
                            ibis.schema(
                                dict(
                                    x="array<int64>",
                                    y="array<string>",
                                    z="array<float64>",
                                    grouper="string",
                                    scalar_column="float64",
                                )
                            )
                        ),
                    ),
                )
            )

            futures.append(
                e.submit(
                    make_job,
                    client.load_table_from_file,
                    io.BytesIO(
                        self.data_dir.joinpath("avro", "struct_table.avro").read_bytes()
                    ),
                    bq.TableReference(testing_dataset, "struct_table"),
                    job_config=bq.LoadJobConfig(
                        write_disposition=write_disposition,
                        source_format=bq.SourceFormat.AVRO,
                    ),
                )
            )

            futures.append(
                e.submit(
                    make_job,
                    client.load_table_from_file,
                    io.StringIO(
                        "\n".join(  # noqa: FLY002
                            [
                                """{"string_col": "1st value", "numeric_col": 0.999999999}""",
                                """{"string_col": "2nd value", "numeric_col": 0.000000002}""",
                            ]
                        )
                    ),
                    bq.TableReference(testing_dataset, "numeric_table"),
                    job_config=bq.LoadJobConfig(
                        write_disposition=write_disposition,
                        schema=BigQuerySchema.from_ibis(
                            ibis.schema(
                                dict(string_col="string", numeric_col="decimal(38, 9)")
                            )
                        ),
                        source_format=bq.SourceFormat.NEWLINE_DELIMITED_JSON,
                    ),
                )
            )

            futures.append(
                e.submit(
                    make_job,
                    client.load_table_from_dataframe,
                    win,
                    bq.TableReference(testing_dataset, "win"),
                    job_config=bq.LoadJobConfig(
                        write_disposition=write_disposition,
                        schema=BigQuerySchema.from_ibis(
                            ibis.schema(dict(g="string", x="!int64", y="int64"))
                        ),
                    ),
                )
            )

            futures.append(
                e.submit(
                    make_job,
                    client.load_table_from_file,
                    io.StringIO("\n".join(f'{{"js": {row}}}' for row in json_types.js)),
                    bq.TableReference(testing_dataset, "json_t"),
                    job_config=bq.LoadJobConfig(
                        write_disposition=write_disposition,
                        schema=BigQuerySchema.from_ibis(ibis.schema(dict(js="json"))),
                        source_format=bq.SourceFormat.NEWLINE_DELIMITED_JSON,
                    ),
                )
            )

            futures.extend(
                e.submit(
                    make_job,
                    client.load_table_from_file,
                    io.BytesIO(
                        self.data_dir.joinpath(
                            "parquet", f"{table}.parquet"
                        ).read_bytes()
                    ),
                    bq.TableReference(testing_dataset, table),
                    job_config=bq.LoadJobConfig(
                        schema=BigQuerySchema.from_ibis(ibis.schema(schema)),
                        write_disposition=write_disposition,
                        source_format=bq.SourceFormat.PARQUET,
                    ),
                )
                for table, schema in TEST_TABLES.items()
            )

            # Test regional endpoints with non-US data.

            futures.extend(
                e.submit(
                    make_job,
                    client.load_table_from_file,
                    io.BytesIO(
                        self.data_dir.joinpath(
                            "parquet", f"{table}.parquet"
                        ).read_bytes()
                    ),
                    bq.TableReference(testing_dataset_tokyo, table),
                    job_config=bq.LoadJobConfig(
                        schema=BigQuerySchema.from_ibis(ibis.schema(schema)),
                        write_disposition=write_disposition,
                        source_format=bq.SourceFormat.PARQUET,
                    ),
                )
                for table, schema in TEST_TABLES.items()
            )

            for fut in concurrent.futures.as_completed(futures):
                fut.result()

    @staticmethod
    def connect(*, tmpdir, worker_id, **kw) -> Backend:
        """Connect to the test project and dataset."""
        credentials, default_project_id = google.auth.default(
            scopes=EXTERNAL_DATA_SCOPES
        )

        project_id = (
            os.environ.get(PROJECT_ID_ENV_VAR, default_project_id) or DEFAULT_PROJECT_ID
        )
        con = ibis.bigquery.connect(
            project_id=project_id, dataset_id=DATASET_ID, credentials=credentials, **kw
        )
        expr = ibis.literal(1)
        try:
            con.execute(expr)
        except gexc.Forbidden:
            pytest.skip(
                f"User does not have access to execute queries against BigQuery project: {project_id}"
            )
        else:
            return con



from __future__ import annotations

from public import public

from ibis.expr.types.generic import Column, Scalar, Value


@public
class MACADDRValue(Value):
    pass


@public
class MACADDRScalar(Scalar, MACADDRValue):
    pass


@public
class MACADDRColumn(Column, MACADDRValue):
    pass


@public
class INETValue(Value):
    pass


@public
class INETScalar(Scalar, INETValue):
    pass


@public
class INETColumn(Column, INETValue):
    pass



from __future__ import annotations

import copy
import functools
import inspect
import itertools
import os
import string

import numpy as np
import pandas as pd
import pytest
from packaging.version import parse as vparse

import ibis
import ibis.expr.datatypes as dt
import ibis.expr.operations as ops
import ibis.expr.types as ir
from ibis.backends.base import _get_backend_names

pytestmark = pytest.mark.benchmark


def make_t():
    return ibis.table(
        [
            ("_timestamp", "int32"),
            ("dim1", "int32"),
            ("dim2", "int32"),
            ("valid_seconds", "int32"),
            ("meas1", "int32"),
            ("meas2", "int32"),
            ("year", "int32"),
            ("month", "int32"),
            ("day", "int32"),
            ("hour", "int32"),
            ("minute", "int32"),
        ],
        name="t",
    )


@pytest.fixture(scope="module")
def t():
    return make_t()


def make_base(t):
    return t[
        (
            (t.year > 2016)
            | ((t.year == 2016) & (t.month > 6))
            | ((t.year == 2016) & (t.month == 6) & (t.day > 6))
            | ((t.year == 2016) & (t.month == 6) & (t.day == 6) & (t.hour > 6))
            | (
                (t.year == 2016)
                & (t.month == 6)
                & (t.day == 6)
                & (t.hour == 6)
                & (t.minute >= 5)
            )
        )
        & (
            (t.year < 2016)
            | ((t.year == 2016) & (t.month < 6))
            | ((t.year == 2016) & (t.month == 6) & (t.day < 6))
            | ((t.year == 2016) & (t.month == 6) & (t.day == 6) & (t.hour < 6))
            | (
                (t.year == 2016)
                & (t.month == 6)
                & (t.day == 6)
                & (t.hour == 6)
                & (t.minute <= 5)
            )
        )
    ]


@pytest.fixture(scope="module")
def base(t):
    return make_base(t)


def make_large_expr(base):
    src_table = base
    src_table = src_table.mutate(
        _timestamp=(src_table["_timestamp"] - src_table["_timestamp"] % 3600)
        .cast("int32")
        .name("_timestamp"),
        valid_seconds=300,
    )

    aggs = []
    for meas in ["meas1", "meas2"]:
        aggs.append(src_table[meas].sum().cast("float").name(meas))
    src_table = src_table.aggregate(
        aggs, by=["_timestamp", "dim1", "dim2", "valid_seconds"]
    )

    part_keys = ["year", "month", "day", "hour", "minute"]
    ts_col = src_table["_timestamp"].cast("timestamp")
    new_cols = {}
    for part_key in part_keys:
        part_col = getattr(ts_col, part_key)()
        new_cols[part_key] = part_col
    src_table = src_table.mutate(**new_cols)
    return src_table[
        [
            "_timestamp",
            "dim1",
            "dim2",
            "meas1",
            "meas2",
            "year",
            "month",
            "day",
            "hour",
            "minute",
        ]
    ]


@pytest.fixture(scope="module")
def large_expr(base):
    return make_large_expr(base)


@pytest.mark.benchmark(group="construction")
@pytest.mark.parametrize(
    "construction_fn",
    [
        pytest.param(lambda *_: make_t(), id="small"),
        pytest.param(lambda t, *_: make_base(t), id="medium"),
        pytest.param(lambda _, base: make_large_expr(base), id="large"),
    ],
)
def test_construction(benchmark, construction_fn, t, base):
    benchmark(construction_fn, t, base)


@pytest.mark.benchmark(group="builtins")
@pytest.mark.parametrize(
    "expr_fn",
    [
        pytest.param(lambda t, _base, _large_expr: t, id="small"),
        pytest.param(lambda _t, base, _large_expr: base, id="medium"),
        pytest.param(lambda _t, _base, large_expr: large_expr, id="large"),
    ],
)
@pytest.mark.parametrize("builtin", [hash, str])
def test_builtins(benchmark, expr_fn, builtin, t, base, large_expr):
    expr = expr_fn(t, base, large_expr)
    benchmark(builtin, expr)


_backends = _get_backend_names(exclude=("pandas",))

_XFAIL_COMPILE_BACKENDS = ("dask", "polars")


@pytest.mark.benchmark(group="compilation")
@pytest.mark.parametrize(
    "module",
    [
        pytest.param(
            mod,
            marks=pytest.mark.xfail(
                condition=mod in _XFAIL_COMPILE_BACKENDS,
                reason=f"{mod} backend doesn't support compiling UnboundTable",
            ),
        )
        for mod in _backends
    ],
)
@pytest.mark.parametrize(
    "expr_fn",
    [
        pytest.param(lambda t, _base, _large_expr: t, id="small"),
        pytest.param(lambda _t, base, _large_expr: base, id="medium"),
        pytest.param(lambda _t, _base, large_expr: large_expr, id="large"),
    ],
)
def test_compile(benchmark, module, expr_fn, t, base, large_expr):
    try:
        mod = getattr(ibis, module)
    except (AttributeError, ImportError) as e:
        pytest.skip(str(e))
    else:
        expr = expr_fn(t, base, large_expr)
        try:
            benchmark(mod.compile, expr)
        except ImportError as e:  # delayed imports
            pytest.skip(str(e))


@pytest.fixture(scope="module")
def pt():
    n = 60_000
    data = pd.DataFrame(
        {
            "key": np.random.choice(16000, size=n),
            "low_card_key": np.random.choice(30, size=n),
            "value": np.random.rand(n),
            "timestamps": pd.date_range(
                start="2023-05-05 16:37:57", periods=n, freq="s"
            ).values,
            "timestamp_strings": pd.date_range(
                start="2023-05-05 16:37:39", periods=n, freq="s"
            ).values.astype(str),
            "repeated_timestamps": pd.date_range(start="2018-09-01", periods=30).repeat(
                int(n / 30)
            ),
        }
    )

    con = ibis.duckdb.connect()
    return con.create_table("df", data)


def high_card_group_by(t):
    return t.group_by(t.key).aggregate(avg_value=t.value.mean())


def cast_to_dates(t):
    return t.timestamps.cast(dt.date)


def cast_to_dates_from_strings(t):
    return t.timestamp_strings.cast(dt.date)


def multikey_group_by_with_mutate(t):
    return (
        t.mutate(dates=t.timestamps.cast("date"))
        .group_by(["low_card_key", "dates"])
        .aggregate(avg_value=lambda t: t.value.mean())
    )


def simple_sort(t):
    return t.order_by([t.key])


def simple_sort_projection(t):
    return t[["key", "value"]].order_by(["key"])


def multikey_sort(t):
    return t.order_by(["low_card_key", "key"])


def multikey_sort_projection(t):
    return t[["low_card_key", "key", "value"]].order_by(["low_card_key", "key"])


def low_card_rolling_window(t):
    return ibis.trailing_range_window(
        ibis.interval(days=2),
        order_by=t.repeated_timestamps,
        group_by=t.low_card_key,
    )


def low_card_grouped_rolling(t):
    return t.value.mean().over(low_card_rolling_window(t))


def high_card_rolling_window(t):
    return ibis.trailing_range_window(
        ibis.interval(days=2),
        order_by=t.repeated_timestamps,
        group_by=t.key,
    )


def high_card_grouped_rolling(t):
    return t.value.mean().over(high_card_rolling_window(t))


def low_card_window(t):
    return ibis.window(group_by=t.low_card_key)


def high_card_window(t):
    return ibis.window(group_by=t.key)


broken_pandas_grouped_rolling = pytest.mark.xfail(
    condition=vparse("1.4") <= vparse(pd.__version__) < vparse("1.4.2"),
    raises=ValueError,
    reason="https://github.com/pandas-dev/pandas/pull/44068",
)


@pytest.mark.benchmark(group="execution")
@pytest.mark.parametrize(
    "expression_fn",
    [
        pytest.param(high_card_group_by, id="high_card_group_by"),
        pytest.param(cast_to_dates, id="cast_to_dates"),
        pytest.param(cast_to_dates_from_strings, id="cast_to_dates_from_strings"),
        pytest.param(multikey_group_by_with_mutate, id="multikey_group_by_with_mutate"),
        pytest.param(simple_sort, id="simple_sort"),
        pytest.param(simple_sort_projection, id="simple_sort_projection"),
        pytest.param(multikey_sort, id="multikey_sort"),
        pytest.param(multikey_sort_projection, id="multikey_sort_projection"),
        pytest.param(
            low_card_grouped_rolling,
            id="low_card_grouped_rolling",
            marks=[broken_pandas_grouped_rolling],
        ),
        pytest.param(
            high_card_grouped_rolling,
            id="high_card_grouped_rolling",
            marks=[broken_pandas_grouped_rolling],
        ),
    ],
)
def test_execute(benchmark, expression_fn, pt):
    expr = expression_fn(pt)
    benchmark(expr.execute)


@pytest.fixture(scope="module")
def part():
    return ibis.table(
        dict(
            p_partkey="int64",
            p_size="int64",
            p_type="string",
            p_mfgr="string",
        ),
        name="part",
    )


@pytest.fixture(scope="module")
def supplier():
    return ibis.table(
        dict(
            s_suppkey="int64",
            s_nationkey="int64",
            s_name="string",
            s_acctbal="decimal(15, 3)",
            s_address="string",
            s_phone="string",
            s_comment="string",
        ),
        name="supplier",
    )


@pytest.fixture(scope="module")
def partsupp():
    return ibis.table(
        dict(
            ps_partkey="int64",
            ps_suppkey="int64",
            ps_supplycost="decimal(15, 3)",
        ),
        name="partsupp",
    )


@pytest.fixture(scope="module")
def nation():
    return ibis.table(
        dict(n_nationkey="int64", n_regionkey="int64", n_name="string"),
        name="nation",
    )


@pytest.fixture(scope="module")
def region():
    return ibis.table(dict(r_regionkey="int64", r_name="string"), name="region")


@pytest.fixture(scope="module")
def tpc_h02(part, supplier, partsupp, nation, region):
    REGION = "EUROPE"
    SIZE = 25
    TYPE = "BRASS"

    expr = (
        part.join(partsupp, part.p_partkey == partsupp.ps_partkey)
        .join(supplier, supplier.s_suppkey == partsupp.ps_suppkey)
        .join(nation, supplier.s_nationkey == nation.n_nationkey)
        .join(region, nation.n_regionkey == region.r_regionkey)
    )

    subexpr = (
        partsupp.join(supplier, supplier.s_suppkey == partsupp.ps_suppkey)
        .join(nation, supplier.s_nationkey == nation.n_nationkey)
        .join(region, nation.n_regionkey == region.r_regionkey)
    )

    subexpr = subexpr[
        (subexpr.r_name == REGION) & (expr.p_partkey == subexpr.ps_partkey)
    ]

    filters = [
        expr.p_size == SIZE,
        expr.p_type.like(f"%{TYPE}"),
        expr.r_name == REGION,
        expr.ps_supplycost == subexpr.ps_supplycost.min(),
    ]
    q = expr.filter(filters)

    q = q.select(
        [
            q.s_acctbal,
            q.s_name,
            q.n_name,
            q.p_partkey,
            q.p_mfgr,
            q.s_address,
            q.s_phone,
            q.s_comment,
        ]
    )

    return q.order_by(
        [
            ibis.desc(q.s_acctbal),
            q.n_name,
            q.s_name,
            q.p_partkey,
        ]
    ).limit(100)


@pytest.mark.benchmark(group="repr")
def test_repr_tpc_h02(benchmark, tpc_h02):
    benchmark(repr, tpc_h02)


@pytest.mark.benchmark(group="repr")
def test_repr_huge_union(benchmark):
    n = 10
    raw_types = [
        "int64",
        "float64",
        "string",
        "array<struct<a: array<string>, b: map<string, array<int64>>>>",
    ]
    tables = [
        ibis.table(
            list(zip(string.ascii_letters, itertools.cycle(raw_types))),
            name=f"t{i:d}",
        )
        for i in range(n)
    ]
    expr = functools.reduce(ir.Table.union, tables)
    benchmark(repr, expr)


@pytest.mark.benchmark(group="node_args")
def test_op_argnames(benchmark):
    t = ibis.table([("a", "int64")])
    expr = t[["a"]]
    benchmark(lambda op: op.argnames, expr.op())


@pytest.mark.benchmark(group="node_args")
def test_op_args(benchmark):
    t = ibis.table([("a", "int64")])
    expr = t[["a"]]
    benchmark(lambda op: op.args, expr.op())


@pytest.mark.benchmark(group="datatype")
def test_complex_datatype_parse(benchmark):
    type_str = "array<struct<a: array<string>, b: map<string, array<int64>>>>"
    expected = dt.Array(
        dt.Struct(dict(a=dt.Array(dt.string), b=dt.Map(dt.string, dt.Array(dt.int64))))
    )
    assert dt.parse(type_str) == expected
    benchmark(dt.parse, type_str)


@pytest.mark.benchmark(group="datatype")
@pytest.mark.parametrize("func", [str, hash])
def test_complex_datatype_builtins(benchmark, func):
    datatype = dt.Array(
        dt.Struct(dict(a=dt.Array(dt.string), b=dt.Map(dt.string, dt.Array(dt.int64))))
    )
    benchmark(func, datatype)


@pytest.mark.benchmark(group="equality")
def test_large_expr_equals(benchmark, tpc_h02):
    benchmark(ir.Expr.equals, tpc_h02, copy.deepcopy(tpc_h02))


@pytest.mark.benchmark(group="datatype")
@pytest.mark.parametrize(
    "dtypes",
    [
        pytest.param(
            [
                obj
                for _, obj in inspect.getmembers(
                    dt,
                    lambda obj: isinstance(obj, dt.DataType),
                )
            ],
            id="singletons",
        ),
        pytest.param(
            dt.Array(
                dt.Struct(
                    dict(
                        a=dt.Array(dt.string),
                        b=dt.Map(dt.string, dt.Array(dt.int64)),
                    )
                )
            ),
            id="complex",
        ),
    ],
)
def test_eq_datatypes(benchmark, dtypes):
    def eq(a, b):
        assert a == b

    benchmark(eq, dtypes, copy.deepcopy(dtypes))


def multiple_joins(table, num_joins):
    for _ in range(num_joins):
        table = table.mutate(dummy=ibis.literal(""))
        table = table.left_join(table, ["dummy"])[[table]]


@pytest.mark.parametrize("num_joins", [1, 10])
@pytest.mark.parametrize("num_columns", [1, 10, 100])
def test_multiple_joins(benchmark, num_joins, num_columns):
    table = ibis.table(
        {f"col_{i:d}": "string" for i in range(num_columns)},
        name="t",
    )
    benchmark(multiple_joins, table, num_joins)


@pytest.fixture
def customers():
    return ibis.table(
        dict(
            customerid="int32",
            name="string",
            address="string",
            citystatezip="string",
            birthdate="date",
            phone="string",
            timezone="string",
            lat="float64",
            long="float64",
        ),
        name="customers",
    )


@pytest.fixture
def orders():
    return ibis.table(
        dict(
            orderid="int32",
            customerid="int32",
            ordered="timestamp",
            shipped="timestamp",
            items="string",
            total="float64",
        ),
        name="orders",
    )


@pytest.fixture
def orders_items():
    return ibis.table(
        dict(orderid="int32", sku="string", qty="int32", unit_price="float64"),
        name="orders_items",
    )


@pytest.fixture
def products():
    return ibis.table(
        dict(
            sku="string",
            desc="string",
            weight_kg="float64",
            cost="float64",
            dims_cm="string",
        ),
        name="products",
    )


@pytest.mark.benchmark(group="compilation")
@pytest.mark.parametrize(
    "module",
    [
        pytest.param(
            mod,
            marks=pytest.mark.xfail(
                condition=mod in _XFAIL_COMPILE_BACKENDS,
                reason=f"{mod} backend doesn't support compiling UnboundTable",
            ),
        )
        for mod in _backends
    ],
)
def test_compile_with_drops(
    benchmark, module, customers, orders, orders_items, products
):
    expr = (
        customers.join(orders, "customerid")
        .join(orders_items, "orderid")
        .join(products, "sku")
        .drop("customerid", "qty", "total", "items")
        .drop("dims_cm", "cost")
        .mutate(o_date=lambda t: t.shipped.date())
        .filter(lambda t: t.ordered == t.shipped)
    )

    try:
        mod = getattr(ibis, module)
    except (AttributeError, ImportError) as e:
        pytest.skip(str(e))
    else:
        benchmark(mod.compile, expr)


def test_repr_join(benchmark, customers, orders, orders_items, products):
    expr = (
        customers.join(orders, "customerid")
        .join(orders_items, "orderid")
        .join(products, "sku")
        .drop("customerid", "qty", "total", "items")
    )
    op = expr.op()
    benchmark(repr, op)


@pytest.mark.parametrize("overwrite", [True, False], ids=["overwrite", "no_overwrite"])
def test_insert_duckdb(benchmark, overwrite, tmp_path):
    pytest.importorskip("duckdb")

    n_rows = int(1e4)
    table_name = "t"
    schema = ibis.schema(dict(a="int64", b="int64", c="int64"))
    t = ibis.memtable(dict.fromkeys(list("abc"), range(n_rows)), schema=schema)

    con = ibis.duckdb.connect(tmp_path / "test_insert.ddb")
    con.create_table(table_name, schema=schema)
    benchmark(con.insert, table_name, t, overwrite=overwrite)


def test_snowflake_medium_sized_to_pandas(benchmark):
    pytest.importorskip("snowflake.connector")

    if (url := os.environ.get("SNOWFLAKE_URL")) is None:
        pytest.skip("SNOWFLAKE_URL environment variable not set")

    con = ibis.connect(url)

    # LINEITEM at scale factor 1 is around 6MM rows, but we limit to 1,000,000
    # to make the benchmark fast enough for development, yet large enough to show a
    # difference if there's a performance hit
    lineitem = con.table("LINEITEM", schema="SNOWFLAKE_SAMPLE_DATA.TPCH_SF1").limit(
        1_000_000
    )

    benchmark.pedantic(lineitem.to_pandas, rounds=5, iterations=1, warmup_rounds=1)


def test_parse_many_duckdb_types(benchmark):
    from ibis.backends.base.sqlglot.datatypes import DuckDBType

    def parse_many(types):
        list(map(DuckDBType.from_string, types))

    types = ["VARCHAR", "INTEGER", "DOUBLE", "BIGINT"] * 1000
    benchmark(parse_many, types)


@pytest.fixture(scope="session")
def sql() -> str:
    return """
    SELECT t1.id as t1_id, x, t2.id as t2_id, y
    FROM t1 INNER JOIN t2
      ON t1.id = t2.id
    """


@pytest.fixture(scope="session")
def ddb(tmp_path_factory):
    duckdb = pytest.importorskip("duckdb")

    N = 20_000_000

    con = duckdb.connect()

    path = str(tmp_path_factory.mktemp("duckdb") / "data.ddb")
    sql = (
        lambda var, table, n=N: f"""
        CREATE TABLE {table} AS
        SELECT ROW_NUMBER() OVER () AS id, {var}
        FROM (
            SELECT {var}
            FROM RANGE({n}) _ ({var})
            ORDER BY RANDOM()
        )
        """
    )

    with duckdb.connect(path) as con:
        con.execute(sql("x", table="t1"))
        con.execute(sql("y", table="t2"))
    return path


def test_duckdb_to_pyarrow(benchmark, sql, ddb) -> None:
    # yes, we're benchmarking duckdb here, not ibis
    #
    # we do this to get a baseline for comparison
    duckdb = pytest.importorskip("duckdb")
    con = duckdb.connect(ddb, read_only=True)

    benchmark(lambda sql: con.sql(sql).to_arrow_table(), sql)


def test_ibis_duckdb_to_pyarrow(benchmark, sql, ddb) -> None:
    pytest.importorskip("duckdb")

    con = ibis.duckdb.connect(ddb, read_only=True)

    expr = con.sql(sql)
    benchmark(expr.to_pyarrow)


@pytest.fixture
def diffs():
    return ibis.table(
        {
            "id": "int64",
            "validation_name": "string",
            "difference": "float64",
            "pct_difference": "float64",
            "pct_threshold": "float64",
            "validation_status": "string",
        },
        name="diffs",
    )


@pytest.fixture
def srcs():
    return ibis.table(
        {
            "id": "int64",
            "validation_name": "string",
            "validation_type": "string",
            "aggregation_type": "string",
            "table_name": "string",
            "column_name": "string",
            "primary_keys": "string",
            "num_random_rows": "string",
            "agg_value": "float64",
        },
        name="srcs",
    )


@pytest.fixture
def nrels():
    return 225


def make_big_union(t, nrels):
    return ibis.union(*[t] * nrels)


@pytest.fixture
def src(srcs, nrels):
    return make_big_union(srcs, nrels)


@pytest.fixture
def diff(diffs, nrels):
    return make_big_union(diffs, nrels)


def test_big_eq_expr(benchmark, src, diff):
    benchmark(ops.core.Node.equals, src.op(), diff.op())


def test_big_join_expr(benchmark, src, diff):
    benchmark(ir.Table.join, src, diff, ["validation_name"], how="outer")


def test_big_join_execute(benchmark, nrels):
    pytest.importorskip("duckdb")

    con = ibis.duckdb.connect()

    # cache to avoid a request-per-union operand
    src = make_big_union(
        con.read_csv(
            "https://github.com/ibis-project/ibis/files/12580336/source_pivot.csv"
        )
        .rename(id="column0")
        .cache(),
        nrels,
    )

    diff = make_big_union(
        con.read_csv(
            "https://github.com/ibis-project/ibis/files/12580340/differences_pivot.csv"
        )
        .rename(id="column0")
        .cache(),
        nrels,
    )

    expr = src.join(diff, ["validation_name"], how="outer")
    t = benchmark.pedantic(expr.to_pyarrow, rounds=1, iterations=1, warmup_rounds=1)
    assert len(t)



from __future__ import annotations

import pytest

import ibis
import ibis.expr.datatypes as dt
from ibis import udf

DB_TYPES = [
    # Exact numbers
    ("BIGINT", dt.int64),
    ("BIT", dt.boolean),
    ("DECIMAL", dt.Decimal(precision=18, scale=0)),
    ("DECIMAL(5, 2)", dt.Decimal(precision=5, scale=2)),
    ("INT", dt.int32),
    ("MONEY", dt.int64),
    ("NUMERIC", dt.Decimal(18, 0)),
    ("NUMERIC(10,5)", dt.Decimal(10, 5)),
    ("NUMERIC(14,3)", dt.Decimal(14, 3)),
    ("SMALLINT", dt.int16),
    ("SMALLMONEY", dt.int32),
    ("TINYINT", dt.int8),
    # Approximate numerics
    ("REAL", dt.float32),
    ("FLOAT", dt.float64),
    ("FLOAT(3)", dt.float32),
    ("FLOAT(25)", dt.float64),
    ("FLOAT(37)", dt.float64),
    # Date and time
    ("DATE", dt.date),
    ("TIME", dt.time),
    ("DATETIME2", dt.timestamp(scale=7)),
    ("DATETIMEOFFSET", dt.timestamp(scale=7, timezone="UTC")),
    ("SMALLDATETIME", dt.Timestamp(scale=0)),
    ("DATETIME", dt.Timestamp(scale=3)),
    # Characters strings
    ("CHAR", dt.string),
    ("TEXT", dt.string),
    ("VARCHAR", dt.string),
    # Unicode character strings
    ("NCHAR", dt.string),
    ("NTEXT", dt.string),
    ("NVARCHAR", dt.string),
    # Binary strings
    ("BINARY", dt.binary),
    ("VARBINARY", dt.binary),
    ("IMAGE", dt.binary),
    # Other data types
    ("UNIQUEIDENTIFIER", dt.uuid),
    ("TIMESTAMP", dt.binary(nullable=False)),
    ("DATETIME2(4)", dt.timestamp(scale=4)),
    ("DATETIMEOFFSET(5)", dt.timestamp(scale=5, timezone="UTC")),
    ("GEOMETRY", dt.geometry),
    ("GEOGRAPHY", dt.geography),
    ("HIERARCHYID", dt.string),
]


@pytest.mark.parametrize(("server_type", "expected_type"), DB_TYPES, ids=str)
def test_get_schema(con, server_type, expected_type, temp_table):
    with con.begin() as c:
        c.execute(f"CREATE TABLE [{temp_table}] (x {server_type})")

    expected_schema = ibis.schema(dict(x=expected_type))

    assert con.get_schema(temp_table) == expected_schema
    assert con.table(temp_table).schema() == expected_schema
    assert con.sql(f"SELECT * FROM [{temp_table}]").schema() == expected_schema


def test_builtin_scalar_udf(con):
    @udf.scalar.builtin
    def difference(a: str, b: str) -> int:
        """Soundex difference between two strings."""

    expr = difference("foo", "moo")
    result = con.execute(expr)
    assert result == 3


def test_builtin_agg_udf(con):
    @udf.agg.builtin
    def count_big(x) -> int:
        """The biggest of counts."""

    ft = con.tables.functional_alltypes
    expr = count_big(ft.id)
    assert expr.execute() == ft.count().execute()


def test_builtin_agg_udf_filtered(con):
    @udf.agg.builtin
    def count_big(x, where: bool = True) -> int:
        """The biggest of counts."""

    ft = con.tables.functional_alltypes
    expr = count_big(ft.id)

    expr = count_big(ft.id, where=ft.id == 1)
    assert expr.execute() == ft[ft.id == 1].count().execute()


@pytest.mark.parametrize("string", ["a", " ", "a ", " a", ""])
def test_glorious_length_function_hack(con, string):
    """Test that the length function works as expected.

    Why wouldn't it, you ask?

    https://learn.microsoft.com/en-us/sql/t-sql/functions/len-transact-sql?view=sql-server-ver16#remarks
    """
    lit = ibis.literal(string)
    expr = lit.length()
    result = con.execute(expr)
    assert result == len(string)



from __future__ import annotations

import datetime
import decimal

import pandas as pd
import pytest
from rich.console import Console

import ibis
import ibis.expr.datatypes as dt
from ibis.expr.types.pretty import format_column, format_values

null = "NULL"


def test_format_int_column():
    values = [None, 1, 2, 3, None]
    fmts, min_len, max_len = format_column(dt.int64, values)
    strs = [str(f) for f in fmts]
    assert strs == [null, "1", "2", "3", null]
    assert min_len == 4
    assert max_len == 4


def test_format_bool_column():
    values = [None, True, False]
    fmts, _, _ = format_column(dt.bool, values)
    strs = [str(f) for f in fmts]
    assert strs == [null, "True", "False"]


def test_format_float_column():
    values = [float("nan"), 1.52, -2.0, 0.0, float("inf"), float("-inf"), None]
    fmts, _, _ = format_column(dt.float64, values)
    strs = [str(f) for f in fmts]
    # matching trailing zeros are stripped
    assert strs == ["nan", "1.52", "-2.00", "0.00", "inf", "-inf", null]


def test_format_big_float_column():
    values = [1.52e6, -2.0e-6]
    fmts, _, _ = format_column(dt.float64, values)
    strs = [str(f) for f in fmts]
    assert strs == ["1.520000e+06", "-2.000000e-06"]


@pytest.mark.parametrize("is_float", [False, True])
def test_format_decimal(is_float):
    values = [decimal.Decimal("1.500"), decimal.Decimal("2.510")]
    if is_float:
        values = [float(v) for v in values]

    # With a scale, render using specified scale, even if backend
    # doesn't return results as a `Decimal` object.
    fmts, _, _ = format_column(dt.Decimal(scale=3), values)
    strs = [str(f) for f in fmts]
    assert strs == ["1.500", "2.510"]

    # Without scale, decimals render same as floats
    fmts, _, _ = format_column(dt.Decimal(scale=None), values)
    strs = [str(f) for f in fmts]
    assert strs == ["1.50", "2.51"]


@pytest.mark.parametrize(
    "t, prec",
    [
        ("2022-02-02 01:02:03", "seconds"),
        ("2022-02-02 01:02:03.123", "milliseconds"),
        ("2022-02-02 01:02:03.123456", "microseconds"),
    ],
)
def test_format_timestamp_column(t, prec):
    a = pd.Timestamp("2022-02-02")
    b = pd.Timestamp(t)
    NaT = pd.Timestamp("NaT")
    fmts, _, _ = format_column(dt.timestamp, [a, b, NaT])
    strs = [str(f) for f in fmts]
    assert strs == [
        a.isoformat(sep=" ", timespec=prec),
        b.isoformat(sep=" ", timespec=prec),
        null,
    ]


@pytest.mark.parametrize(
    "t, prec",
    [
        ("01:02:03", "seconds"),
        ("01:02:03.123", "milliseconds"),
        ("01:02:03.123456", "microseconds"),
    ],
)
def test_format_time_column(t, prec):
    a = datetime.time(4, 5, 6)
    b = datetime.time.fromisoformat(t)
    NaT = pd.Timestamp("NaT")
    fmts, _, _ = format_column(dt.time, [a, b, NaT])
    strs = [str(f) for f in fmts]
    assert strs == [
        a.isoformat(timespec=prec),
        b.isoformat(timespec=prec),
        null,
    ]


@pytest.mark.parametrize("is_date", [True, False])
def test_format_date_column(is_date):
    cls = datetime.date if is_date else datetime.datetime
    values = [cls(2022, 1, 1), cls(2022, 1, 2)]
    fmts, _, _ = format_column(dt.date, values)
    strs = [str(f) for f in fmts]
    assert strs == ["2022-01-01", "2022-01-02"]


def test_format_interval_column():
    values = [datetime.timedelta(seconds=1)]
    fmts, _, _ = format_column(dt.Interval("s"), values)
    strs = [str(f) for f in fmts]
    assert strs == [str(v) for v in values]


def test_format_string_column():
    max_string = ibis.options.repr.interactive.max_string
    values = [None, "", "test\t\r\n\v\f", "a string", "x" * (max_string + 10)]
    fmts, min_len, max_len = format_column(dt.string, values)
    strs = [str(f) for f in fmts]
    assert strs == [
        null,
        "~",
        "test\\t\\r\\n\\v\\f",
        "a string",
        "x" * (max_string - 1) + "…",
    ]
    assert min_len == 20
    assert max_len == max(map(len, strs))


def test_format_short_string_column():
    values = [None, "", "ab", "cd"]
    fmts, min_len, max_len = format_column(dt.string, values)
    strs = [str(f) for f in fmts]
    assert strs == [null, "~", "ab", "cd"]
    assert min_len == 4
    assert max_len == 4


def test_format_struct_column():
    dtype = dt.Struct({"x": "int", "y": "float"})
    values = [{"x": 1, "y": 2.5}, None]
    fmts, min_len, max_len = format_column(dtype, values)
    assert str(fmts[1]) == null
    assert min_len == 20
    assert max_len is None


def test_format_map_column():
    dtype = dt.Map(dt.String(), dt.Int32())
    values = [[("x", 1), ("y", 2.5)], None]
    fmts, min_len, max_len = format_column(dtype, values)
    assert str(fmts[1]) == null
    assert min_len == 20
    assert max_len is None


def test_format_fully_null_column():
    values = [None, None, None]
    fmts, *_ = format_column(dt.int64, values)
    strs = [str(f) for f in fmts]
    assert strs == [null, null, null]


def test_all_empty_groups_repr():
    values = [float("nan"), float("nan")]
    dtype = dt.float64
    fmts, *_ = format_column(dtype, values)
    assert list(map(str, fmts)) == ["nan", "nan"]


def test_non_interactive_column_repr():
    t = ibis.table(dict(names="string", values="int"))
    expr = t.names
    console = Console()
    with console.capture() as capture:
        console.print(expr)

    result = capture.get()
    assert "Selection" not in result


def test_format_link():
    result = format_values(dt.string, ["https://ibis-project.org"])[0]
    assert str(result) == "https://ibis-project.org"
    assert result.spans[0].style == "link https://ibis-project.org"

    # Check that long urls are truncated visually, but the link
    # itself is still the full url.
    long = "https://" + "x" * 1000
    result = format_values(dt.string, [long])[0]
    assert str(result).startswith("https://xxxxx")
    assert len(str(result)) < 1000
    assert result.spans[0].style == "link " + long

    # not links
    results = format_values(dt.string, ["https://", "https:", "https"])
    assert all(not rendered.spans for rendered in results)



from __future__ import annotations

import ibis

from .conftest import tpch_test


@tpch_test
def test_tpc_h09(part, supplier, lineitem, partsupp, orders, nation):
    """Product Type Profit Measure Query (Q9)"""
    COLOR = "green"

    q = lineitem
    q = q.join(supplier, supplier.s_suppkey == lineitem.l_suppkey)
    q = q.join(
        partsupp,
        (partsupp.ps_suppkey == lineitem.l_suppkey)
        & (partsupp.ps_partkey == lineitem.l_partkey),
    )
    q = q.join(part, part.p_partkey == lineitem.l_partkey)
    q = q.join(orders, orders.o_orderkey == lineitem.l_orderkey)
    q = q.join(nation, supplier.s_nationkey == nation.n_nationkey)

    q = q[
        (q.l_extendedprice * (1 - q.l_discount) - q.ps_supplycost * q.l_quantity).name(
            "amount"
        ),
        q.o_orderdate.year().name("o_year"),
        q.n_name.name("nation"),
        q.p_name,
    ]

    q = q.filter([q.p_name.like("%" + COLOR + "%")])

    gq = q.group_by([q.nation, q.o_year])
    q = gq.aggregate(sum_profit=q.amount.sum())
    q = q.order_by([q.nation, ibis.desc(q.o_year)])
    return q



from __future__ import annotations

import pytest
import sqlglot as sg
from pytest import param

import ibis.common.exceptions as com
import ibis.expr.datatypes as dt
from ibis.backends.bigquery.datatypes import BigQueryType


@pytest.mark.parametrize(
    ("datatype", "expected"),
    [
        param(dt.float32, "FLOAT64", id="float32"),
        param(dt.float64, "FLOAT64", id="float64"),
        param(dt.uint8, "INT64", id="uint8"),
        param(dt.uint16, "INT64", id="uint16"),
        param(dt.uint32, "INT64", id="uint32"),
        param(dt.int8, "INT64", id="int8"),
        param(dt.int16, "INT64", id="int16"),
        param(dt.int32, "INT64", id="int32"),
        param(dt.int64, "INT64", id="int64"),
        param(dt.string, "STRING", id="string"),
        param(dt.Array(dt.int64), "ARRAY<INT64>", id="array<int64>"),
        param(dt.Array(dt.string), "ARRAY<STRING>", id="array<string>"),
        param(
            dt.Struct.from_tuples(
                [("a", dt.int64), ("b", dt.string), ("c", dt.Array(dt.string))]
            ),
            "STRUCT<a INT64, b STRING, c ARRAY<STRING>>",
            id="struct",
        ),
        param(dt.date, "DATE", id="date"),
        param(dt.timestamp, "DATETIME", id="datetime"),
        param(
            dt.Timestamp(timezone="UTC"),
            "TIMESTAMP",
            id="timestamp_with_utc_tz",
        ),
        param(
            dt.Timestamp(timezone="US/Eastern"),
            "TIMESTAMP",
            marks=pytest.mark.xfail(
                raises=com.UnsupportedBackendType, reason="Not supported in BigQuery"
            ),
            id="timestamp_with_other_tz",
        ),
        param(
            dt.Array(dt.Struct({"a": dt.string})),
            "ARRAY<STRUCT<a STRING>>",
            id="array<struct>",
        ),
        param(dt.Decimal(38, 9), "NUMERIC", id="decimal-numeric"),
        param(dt.Decimal(76, 38), "BIGNUMERIC", id="decimal-bignumeric"),
        param(
            dt.GeoSpatial(geotype="geography", srid=4326), "GEOGRAPHY", id="geography"
        ),
        param(
            dt.GeoSpatial(geotype="geography"),
            "GEOGRAPHY",
            marks=pytest.mark.xfail(
                raises=com.UnsupportedBackendType,
                reason="Should use the WGS84 reference ellipsoid.",
            ),
            id="geography-no-srid",
        ),
    ],
)
def test_simple(datatype, expected):
    assert BigQueryType.to_string(datatype) == expected


@pytest.mark.parametrize("datatype", [dt.uint64, dt.Decimal(8, 3)])
def test_simple_failure_mode(datatype):
    with pytest.raises(com.UnsupportedBackendType):
        BigQueryType.to_string(datatype)


def test_struct_type():
    dtype = dt.Array(dt.int64)
    parsed_type = sg.parse_one("BIGINT[]", into=sg.exp.DataType, read="duckdb")

    expected = "ARRAY<INT64>"

    assert parsed_type.sql(dialect="bigquery") == expected
    assert BigQueryType.to_string(dtype) == expected



from __future__ import annotations

import decimal
import functools
import math
import operator
from operator import methodcaller

import numpy as np
import pandas as pd
import pytest
from pytest import param

import ibis
import ibis.expr.datatypes as dt
from ibis.backends.dask.tests.conftest import TestConf as tm


@pytest.mark.parametrize(
    "op",
    [
        # comparison
        operator.eq,
        operator.ne,
        operator.lt,
        operator.le,
        operator.gt,
        operator.ge,
    ],
)
def test_binary_operations(t, df, op):
    expr = op(t.plain_float64, t.plain_int64)
    result = expr.execute()
    expected = op(df.plain_float64, df.plain_int64).compute()
    tm.assert_series_equal(
        result.reset_index(drop=True).rename("tmp"),
        expected.reset_index(drop=True).rename("tmp"),
    )


@pytest.mark.parametrize("op", [operator.and_, operator.or_, operator.xor])
def test_binary_boolean_operations(t, pandas_df, op):
    expr = op(t.plain_int64 == 1, t.plain_int64 == 2)
    result = expr.execute()
    expected = op(pandas_df.plain_int64 == 1, pandas_df.plain_int64 == 2)
    tm.assert_series_equal(
        result.reset_index(drop=True),
        expected.reset_index(drop=True),
    )


def operate(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except decimal.InvalidOperation:
            return decimal.Decimal("NaN")

    return wrapper


@pytest.mark.parametrize(
    ("ibis_func", "pandas_func"),
    [
        param(
            methodcaller("round", 2),
            lambda x: x.quantize(decimal.Decimal(".00")),
            id="round_2",
        ),
        param(
            methodcaller("round", 0),
            lambda x: x.quantize(decimal.Decimal("0.")),
            id="round_0",
        ),
        param(methodcaller("ceil"), lambda x: decimal.Decimal(math.ceil(x)), id="ceil"),
        param(
            methodcaller("floor"), lambda x: decimal.Decimal(math.floor(x)), id="floor"
        ),
        param(
            methodcaller("exp"),
            methodcaller("exp"),
            id="exp",
            marks=pytest.mark.xfail(
                reason="Unable to normalize Decimal('2.71513316E+43') as decimal with precision 12 and scale 3",
                raises=TypeError,
            ),
        ),
        param(
            methodcaller("sign"),
            lambda x: x if not x else decimal.Decimal(1).copy_sign(x),
            id="sign",
        ),
        param(methodcaller("sqrt"), operate(lambda x: x.sqrt()), id="sqrt"),
        param(
            methodcaller("log", 2),
            operate(lambda x: x.ln() / decimal.Decimal(2).ln()),
            id="log_2",
        ),
        param(methodcaller("ln"), operate(lambda x: x.ln()), id="ln"),
        param(
            methodcaller("log2"),
            operate(lambda x: x.ln() / decimal.Decimal(2).ln()),
            id="log2",
        ),
        param(methodcaller("log10"), operate(lambda x: x.log10()), id="log10"),
    ],
)
def test_math_functions_decimal(t, pandas_df, ibis_func, pandas_func):
    dtype = dt.Decimal(12, 3)
    context = decimal.Context(prec=dtype.precision)
    p = decimal.Decimal(f"{'0' * (dtype.precision - dtype.scale)}.{'0' * dtype.scale}")

    def func(x):
        x = context.create_decimal(x)
        x = pandas_func(x)
        if math.isnan(x):
            return float("nan")
        return x.quantize(p)

    expr = ibis_func(t.float64_as_strings.cast(dtype))
    result = expr.execute()
    expected = pandas_df.float64_as_strings.map(func, na_action="ignore")
    tm.assert_series_equal(result, expected, check_names=False)


def test_round_decimal_with_negative_places(t):
    type = dt.Decimal(12, 3)
    expr = t.float64_as_strings.cast(type).round(-1)
    result = expr.execute()
    expected = pd.Series(
        list(map(decimal.Decimal, ["1.0E+2", "2.3E+2", "-1.00E+3"])),
        name="float64_as_strings",
    )
    tm.assert_series_equal(result, expected)


@pytest.mark.parametrize(
    ("ibis_func", "dask_func"),
    [
        (
            lambda x: x.quantile([0.25, 0.75]),
            lambda x: list(x.quantile([0.25, 0.75])),
        )
    ],
)
@pytest.mark.parametrize("column", ["float64_with_zeros", "int64_with_zeros"])
def test_quantile_list(t, pandas_df, ibis_func, dask_func, column):
    expr = ibis_func(t[column])
    result = expr.execute()
    expected = dask_func(pandas_df[column])
    assert result == expected


@pytest.mark.parametrize(
    ("ibis_func", "dask_func"),
    [
        (lambda x: x.quantile(0), lambda x: x.quantile(0)),
        (lambda x: x.quantile(1), lambda x: x.quantile(1)),
        (
            lambda x: x.quantile(0.5),
            lambda x: x.quantile(0.5),
        ),
    ],
)
def test_quantile_scalar(t, pandas_df, ibis_func, dask_func):
    result = ibis_func(t.float64_with_zeros).execute()
    expected = dask_func(pandas_df.float64_with_zeros)
    assert result == expected

    result = ibis_func(t.int64_with_zeros).execute()
    expected = dask_func(pandas_df.int64_with_zeros)
    assert result == expected


@pytest.mark.parametrize(
    ("ibis_func", "exc"),
    [
        # no lower/upper specified
        (lambda x: x.clip(), ValueError),
        # out of range on quantile
        (lambda x: x.quantile(5.0), ValueError),
    ],
)
def test_arraylike_functions_transform_errors(t, df, ibis_func, exc):
    with pytest.raises(exc):
        ibis_func(t.float64_with_zeros).execute()


def test_ifelse_returning_bool(con):
    one = ibis.literal(1)
    two = ibis.literal(2)
    true = ibis.literal(True)
    false = ibis.literal(False)
    expr = ibis.ifelse(one + one == two, true, false)
    result = con.execute(expr)
    assert result is np.bool_(True)



from __future__ import annotations

from typing import Callable

import dask.array as da
import dask.dataframe as dd
import numpy as np
import pandas as pd

from ibis.backends.pandas.helpers import PandasUtils


class DaskUtils(PandasUtils):
    @classmethod
    def merge(cls, *args, **kwargs):
        return dd.merge(*args, **kwargs)

    @classmethod
    def merge_asof(cls, *args, **kwargs):
        return dd.merge_asof(*args, **kwargs)

    @classmethod
    def concat(cls, dfs, **kwargs):
        if isinstance(dfs, dict):
            dfs = [v.rename(k) for k, v in dfs.items()]
        return dd.concat(dfs, **kwargs)

    @classmethod
    def asseries(cls, value, like=None):
        """Ensure that value is a pandas Series object, broadcast if necessary."""

        if isinstance(value, dd.Series):
            return value
        elif isinstance(value, dd.core.Scalar):
            # Create a Dask array from the Dask scalar
            try:
                dtype = value.dtype
            except AttributeError:
                #     @property
                #     def dtype(self):
                # >       return self._meta.dtype
                # E       AttributeError: 'Timestamp' object has no attribute 'dtype'
                dtype = object
            array = da.from_delayed(value.to_delayed(), (1,), dtype=dtype)
            # Create a Dask series from the Dask array
            return dd.from_array(array)
        elif isinstance(value, pd.Series):
            return dd.from_pandas(value, npartitions=1)
        elif like is not None:
            if isinstance(value, (tuple, list, dict)):
                fn = lambda df: pd.Series([value] * len(df), index=df.index)
            else:
                fn = lambda df: pd.Series(value, index=df.index)
            return like.map_partitions(fn)
        else:
            return dd.from_pandas(pd.Series([value]), npartitions=1)

    @classmethod
    def asframe(cls, values: dict | tuple):
        # TODO(kszucs): prefer using assign instead of concat
        """Construct a DataFrame from a dict or tuple of Series objects."""
        if isinstance(values, dict):
            names, values = zip(*values.items())
        elif isinstance(values, tuple):
            names = [f"_{i}" for i in range(len(values))]
        else:
            raise TypeError(f"values must be a dict, or tuple; got {type(values)}")

        all_scalars = True
        representative = None
        for v in values:
            if isinstance(v, dd.Series):
                all_scalars = False
                representative = v
                break

        columns = [cls.asseries(v, like=representative) for v in values]
        columns = [v.rename(k) for k, v in zip(names, columns)]

        # dd.concat turns decimal.Decimal("NaN") into np.nan for some reason
        df = dd.concat(columns, axis=1)
        return df, all_scalars

    @classmethod
    def rowwise(cls, func: Callable, operands, name, dtype):
        if dtype == np.dtype("<M8[D]"):
            # Getting the following error otherwise:
            #     TypeError: dtype=datetime64[D] is not supported. Supported
            #     resolutions are 's', 'ms', 'us', and 'ns'
            dtype = "datetime64[s]"
        # dealing with a collection of series objects
        df, _ = cls.asframe(operands)
        return df.apply(func, axis=1, meta=(name, dtype))

    @classmethod
    def elementwise(cls, func, operands, name, dtype):
        meta = (name, dtype)
        value = operands.pop(next(iter(operands)))
        if isinstance(value, dd.Series):
            # dealing with a single series object
            if operands:
                return value.apply(func, **operands, meta=meta)
            else:
                return value.map(func, meta=meta, na_action="ignore")
        else:
            # dealing with a single scalar object
            return func(value, **operands)

    @classmethod
    def partitionwise(cls, func, operands, name, dtype):
        cols = {}
        kwargs = {}
        for name, operand in operands.items():
            if isinstance(operand, (tuple, list)):
                for i, v in enumerate(operand):
                    cols[f"{name}_{i}"] = v
                kwargs[name] = tuple(f"{name}_{i}" for i in range(len(operand)))
            else:
                cols[name] = operand
                kwargs[name] = name

        def mapper(df):
            unpacked = {}
            for k, operand in kwargs.items():
                if isinstance(operand, (tuple, list)):
                    unpacked[k] = [df[col] for col in operand]
                else:
                    unpacked[k] = df[operand]
            return func(df, **unpacked)

        df, _ = cls.asframe(cols)
        return df.map_partitions(mapper, meta=(name, dtype))


def add_globally_consecutive_column(
    df: dd.DataFrame | dd.Series,
    name: str = "_ibis_index",
    set_as_index: bool = True,
) -> dd.DataFrame:
    """Add a column that is globally consecutive across the distributed data.

    By construction, this column is already sorted and can be used to partition
    the data.
    This column can act as if we had a global index across the distributed data.
    This index needs to be consecutive in the range of [0, len(df)), allows
    downstream operations to work properly.
    The default index of dask dataframes is to be consecutive within each partition.

    Important properties:

    - Each row has a unique id (i.e. a value in this column)
    - The global index that's added is consecutive in the same order that the rows currently are in.
    - IDs within each partition are already sorted

    We also do not explicitly deal with overflow in the bounds.

    Parameters
    ----------
    df: dd.DataFrame
        Dataframe to add the column to
    name: str
        Name of the column to use. Default is _ibis_index
    set_as_index: bool
        If True, will set the consecutive column as the index. Default is True.

    Returns
    -------
    dd.DataFrame
        New dask dataframe with sorted partitioned index

    """
    if isinstance(df, dd.Series):
        df = df.to_frame()

    if name in df.columns:
        raise ValueError(f"Column {name} is already present in DataFrame")

    df = df.assign(**{name: 1})
    df = df.assign(**{name: df[name].cumsum() - 1})
    if set_as_index:
        df = df.reset_index(drop=True)
        df = df.set_index(name, sorted=True)

    # No elegant way to rename index https://github.com/dask/dask/issues/4950
    df = df.map_partitions(pd.DataFrame.rename_axis, None, axis="index")

    return df



from __future__ import annotations

import pytest
from pytest import param

import ibis
from ibis import window
from ibis.tests.util import assert_equal

pytest.importorskip("impala")


@pytest.fixture(scope="module")
def alltypes(mockcon):
    return mockcon.table("alltypes")


def assert_sql_equal(expr, snapshot, out="out.sql"):
    result = ibis.to_sql(expr, dialect="impala")
    snapshot.assert_match(result, out)


def test_aggregate_in_projection(alltypes, snapshot):
    t = alltypes
    proj = t[t, (t.f / t.f.sum()).name("normed_f")]
    assert_sql_equal(proj, snapshot)


def test_add_default_order_by(alltypes, snapshot):
    t = alltypes

    first = t.f.first().name("first")
    last = t.f.last().name("last")
    lag = t.f.lag().name("lag")
    diff = (t.f.lead() - t.f).name("fwd_diff")
    lag2 = t.f.lag().over(window(order_by=t.d)).name("lag2")
    grouped = t.group_by("g")
    proj = grouped.mutate([lag, diff, first, last, lag2])
    assert_sql_equal(proj, snapshot)


@pytest.mark.parametrize(
    "window",
    [
        param(window(preceding=0), id="prec_0"),
        param(window(following=0), id="foll_0"),
        param(window(preceding=5), id="prec_5"),
        param(window(preceding=5, following=0), id="prec_5_foll_0"),
        param(window(preceding=5, following=2), id="prec_5_foll_2"),
        param(window(following=2), id="foll_2"),
        param(window(following=2, preceding=0), id="foll_2_prec_0"),
        param(window(following=(5, 10)), id="foll_5_10"),
        param(window(preceding=(10, 5)), id="foll_10_5"),
        param(ibis.cumulative_window(), id="cumulative"),
        param(ibis.trailing_window(10), id="trailing_10"),
    ],
)
def test_window_frame_specs(alltypes, window, snapshot):
    t = alltypes

    w2 = window.order_by(t.f)
    expr = t.select(foo=t.d.sum().over(w2))
    assert_sql_equal(expr, snapshot)


def test_window_rows_with_max_lookback(alltypes):
    mlb = ibis.rows_with_max_lookback(3, ibis.interval(days=3))
    t = alltypes
    w = ibis.trailing_window(mlb, order_by=t.i)
    expr = t.a.sum().over(w)
    with pytest.raises(NotImplementedError):
        ibis.to_sql(expr, dialect="impala")


@pytest.mark.parametrize("name", ["sum", "min", "max", "mean"])
def test_cumulative_functions(alltypes, name, snapshot):
    t = alltypes
    w = ibis.window(order_by=t.d)

    func = getattr(t.f, name)
    cumfunc = getattr(t.f, f"cum{name}")

    expr = cumfunc().over(w).name("foo")
    expected = func().over(ibis.cumulative_window(order_by=t.d)).name("foo")

    expr1 = t.select(expr)
    expr2 = t.select(expected)

    assert_sql_equal(expr1, snapshot, "out1.sql")
    assert_sql_equal(expr2, snapshot, "out2.sql")


def test_nested_analytic_function(alltypes, snapshot):
    t = alltypes

    w = window(order_by=t.f)
    expr = (t.f - t.f.lag()).lag().over(w).name("foo")
    result = t.select(expr)
    assert_sql_equal(result, snapshot)


def test_rank_functions(alltypes, snapshot):
    t = alltypes

    proj = t[t.g, t.f.rank().name("minr"), t.f.dense_rank().name("denser")]
    assert_sql_equal(proj, snapshot)


def test_multiple_windows(alltypes, snapshot):
    t = alltypes

    w = window(group_by=t.g)

    expr = t.f.sum().over(w) - t.f.sum()
    proj = t.select(t.g, result=expr)

    assert_sql_equal(proj, snapshot)


def test_order_by_desc(alltypes, snapshot):
    t = alltypes

    w = window(order_by=ibis.desc(t.f))

    proj = t[t.f, ibis.row_number().over(w).name("revrank")]
    assert_sql_equal(proj, snapshot, "out1.sql")

    expr = t.group_by("g").order_by(ibis.desc(t.f))[t.d.lag().name("foo"), t.a.max()]
    assert_sql_equal(expr, snapshot, "out2.sql")


def test_row_number_does_not_require_order_by(alltypes, snapshot):
    t = alltypes

    expr = t.group_by(t.g).mutate(ibis.row_number().name("foo"))
    assert_sql_equal(expr, snapshot, "out1.sql")

    expr = t.group_by(t.g).order_by(t.f).mutate(ibis.row_number().name("foo"))
    assert_sql_equal(expr, snapshot, "out2.sql")


def test_row_number_properly_composes_with_arithmetic(alltypes, snapshot):
    t = alltypes
    w = ibis.window(order_by=t.f)
    expr = t.mutate(new=ibis.row_number().over(w) / 2)
    assert_sql_equal(expr, snapshot)


def test_propagate_nested_windows(alltypes, snapshot):
    # GH #469
    t = alltypes

    w = ibis.window(group_by=t.g, order_by=t.f)

    col = (t.f - t.f.lag()).lag()

    # propagate down here!
    result = col.over(w)
    ex_expr = (t.f - t.f.lag().over(w)).lag().over(w)
    assert_equal(result, ex_expr)

    expr = t.select(col.over(w).name("foo"))
    assert_sql_equal(expr, snapshot)



import ibis


star1 = ibis.table(
    name="star1",
    schema={"c": "int32", "f": "float64", "foo_id": "string", "bar_id": "string"},
)
star2 = ibis.table(
    name="star2", schema={"foo_id": "string", "value1": "float64", "value3": "float64"}
)
star3 = ibis.table(name="star3", schema={"bar_id": "string", "value2": "float64"})

result = (
    star1.left_join(star2, star1.foo_id == star2.foo_id)
    .inner_join(star3, star1.bar_id == star3.bar_id)
    .select(star1.c, star1.f, star1.foo_id, star1.bar_id, star2.value1, star3.value2)
)



from __future__ import annotations

import dask.dataframe as dd
import pandas as pd
import pandas.api.types as pdt

import ibis.expr.datatypes as dt
from ibis.backends.pandas.convert import PandasConverter
from ibis.formats.pandas import DataMapper, PandasType


class DaskConverter(DataMapper):
    @classmethod
    def convert_scalar(cls, obj, dtype):
        return PandasConverter.convert_scalar(obj, dtype)

    @classmethod
    def convert_column(cls, obj, dtype):
        pandas_type = PandasType.from_ibis(dtype)

        method_name = f"convert_{dtype.__class__.__name__}"
        convert_method = getattr(cls, method_name, cls.convert_default)

        return convert_method(obj, dtype, pandas_type)

    @classmethod
    def convert_default(cls, s, dtype, pandas_type):
        if pandas_type == object:
            func = lambda x: x if x is pd.NA else dt.normalize(dtype, x)
            meta = (s.name, pandas_type)
            return s.map(func, na_action="ignore", meta=meta).astype(pandas_type)
        else:
            return s.astype(pandas_type)

    @classmethod
    def convert_Integer(cls, s, dtype, pandas_type):
        if pdt.is_datetime64_any_dtype(s.dtype):
            return s.astype("int64").floordiv(int(1e9)).astype(pandas_type)
        else:
            return s.astype(pandas_type)

    convert_SignedInteger = convert_UnsignedInteger = convert_Integer
    convert_Int64 = convert_Int32 = convert_Int16 = convert_Int8 = convert_SignedInteger
    convert_UInt64 = (
        convert_UInt32
    ) = convert_UInt16 = convert_UInt8 = convert_UnsignedInteger

    @classmethod
    def convert_Floating(cls, s, dtype, pandas_type):
        if pdt.is_datetime64_any_dtype(s.dtype):
            return s.astype("int64").floordiv(int(1e9)).astype(pandas_type)
        else:
            return s.astype(pandas_type)

    convert_Float64 = convert_Float32 = convert_Float16 = convert_Floating

    @classmethod
    def convert_Timestamp(cls, s, dtype, pandas_type):
        if isinstance(s.dtype, pd.DatetimeTZDtype):
            return s.dt.tz_convert(dtype.timezone)
        elif pdt.is_datetime64_dtype(s.dtype):
            return s.dt.tz_localize(dtype.timezone)
        elif pdt.is_numeric_dtype(s.dtype):
            return dd.to_datetime(s, unit="s").dt.tz_localize(dtype.timezone)
        else:
            return dd.to_datetime(s, utc=True).dt.tz_localize(dtype.timezone)

    @classmethod
    def convert_Date(cls, s, dtype, pandas_type):
        if isinstance(s.dtype, pd.DatetimeTZDtype):
            s = s.dt.tz_convert("UTC").dt.tz_localize(None)
        elif pdt.is_numeric_dtype(s.dtype):
            s = dd.to_datetime(s, unit="D")
        else:
            s = dd.to_datetime(s)

        return s.dt.normalize()

    @classmethod
    def convert_String(cls, s, dtype, pandas_type):
        # TODO(kszucs): should switch to the new pandas string type and convert
        # object columns using s.convert_dtypes() method
        return s.map(str, na_action="ignore").astype(object)
